# pylint: disable=redefined-outer-name
from __future__ import print_function

import os
import shutil
from tempfile import mkdtemp

import pytest
from slash.utils import suite_files


def test_iter_suite_paths_files_abspaths(filename, paths):
    with open(filename, 'w') as f:
        f.write('\n'.join(paths))

    assert [p for p, _ in suite_files.iter_suite_file_paths([filename])] == paths


def test_iter_suite_paths_files_relpath(filename, paths):
    with open(filename, 'w') as f:
        for path in paths:
            relpath = os.path.relpath(path, os.path.dirname(filename))
            assert not os.path.isabs(relpath)
            f.write(relpath)
            f.write('\n')

    assert [p for p, _ in suite_files.iter_suite_file_paths([filename])] == [os.path.abspath(p) for p in paths]


def test_suite_files(suite, suite_test, suite_file):  # pylint: disable=unused-argument
    suite.run(args=[])


def test_suite_file_with_filter(suite, suite_test, suite_file):
    path = suite.commit()
    with suite_file.open('w') as f:
        print(path, '# filter:', suite_test.name, file=f)
    suite.run(args=[])


def test_parse_filter_string():
    suite_entry = suite_files._parse_path_filter_and_repeat('some_path.py # filter: bla')  # pylint: disable=protected-access
    assert suite_entry.path == 'some_path.py'
    assert suite_entry.matcher is not None
    assert suite_entry.matcher.matches('bla')
    assert not suite_entry.matcher.matches('blooop')


@pytest.mark.parametrize("string", [
    'some_path.py #',
    'some_path.py # bbb',
    'some_path.py#'
])
def test_parse_filter_string_no_filter(string):
    suite_entry = suite_files._parse_path_filter_and_repeat(string)  # pylint: disable=protected-access
    assert suite_entry.path == 'some_path.py'
    assert suite_entry.matcher is None


def test_iter_suite_file_paths_nested_filter(tmpdir):

    test_filename = '/some/test/file.py'
    suite_file1 = tmpdir.join('file1.txt')
    suite_file2 = tmpdir.join('file2.txt')

    with suite_file1.open('w') as f:
        print(suite_file2, '# filter: not blue', file=f)

    with suite_file2.open('w') as f:
        print(test_filename, '# filter: green', file=f)

    [(item, matcher)] = suite_files.iter_suite_file_paths([str(suite_file1)])
    assert item == test_filename
    assert matcher.matches('green')
    assert not matcher.matches('blue')
    assert not matcher.matches('green blue')


def test_parse_repeat_string():
    suite_entry = suite_files._parse_path_filter_and_repeat('some_path.py # repeat: 5')  # pylint: disable=protected-access
    assert suite_entry.path == 'some_path.py'
    assert suite_entry.matcher is None
    assert suite_entry.repeat == 5


@pytest.mark.parametrize("string", [
    'some_path.py # filter: bla, repeat: 5',
    'some_path.py # repeat: 5, filter: bla',
])
def test_parse_repeat_string_with_filter(string):
    suite_entry = suite_files._parse_path_filter_and_repeat(string)  # pylint: disable=protected-access
    assert suite_entry.path == 'some_path.py'
    assert suite_entry.matcher is not None
    assert suite_entry.matcher.matches('bla')
    assert not suite_entry.matcher.matches('blooop')
    assert suite_entry.matcher is not None
    assert suite_entry.repeat == 5


@pytest.fixture
def suite_file(tmpdir, suite, suite_test, config_override):
    path = suite.commit()
    suite_file_path = tmpdir.join('suite.txt')
    with suite_file_path.open('w') as f:
        f.write(suite_test.get_full_address(path))
    suite.deselect_all(exclude=[suite_test])
    config_override('run.suite_files', [str(suite_file_path)])
    return suite_file_path


@pytest.mark.parametrize('use_relpath', [True, False])
def test_files_containing_files(filename, paths, use_relpath):
    filename2 = os.path.join(os.path.dirname(filename), 'file2.txt')

    with open(filename2, 'w') as f:
        f.write('\n'.join(paths[::-1]))

    if use_relpath:
        filename2 = os.path.basename(filename2)

    with open(filename, 'w') as f:
        f.write('\n'.join(paths))
        f.write('\n')
        f.write(filename2)

    assert [p for p, _ in suite_files.iter_suite_file_paths([filename])] == paths + paths[::-1]


def test_slash_run_with_suite_file(suite, suite_test, tmpdir):
    path = suite.commit()
    with tmpdir.join('suitefile').open('w') as suite_file:  # pylint: disable=redefined-outer-name
        _fill_suite_file(path, [suite_test], suite_file=suite_file)

    for t in suite:
        if t is not suite_test:
            t.expect_deselect()

    suite.run(args=[], additional_args=['-f', suite_file.name])


def test_slash_run_with_suite_file_invalid_test(suite, suite_test, tmpdir):
    path = suite.commit()

    additional_test = suite[0]
    assert additional_test is not suite_test

    with tmpdir.join('suitefile').open('w') as suite_file:  # pylint: disable=redefined-outer-name
        _fill_suite_file(path, [additional_test], suite_file=suite_file, corrupt=False)
        _fill_suite_file(path, [suite_test], suite_file=suite_file, corrupt=True)

    for t in suite:
        t.expect_deselect()

    summary = suite.run(args=[], additional_args=['-f', suite_file.name], expect_session_errors=True)
    assert summary.exit_code != 0
    assert 'Cannot find test' in summary.get_console_output()
    assert 'CORRUPT' in summary.get_console_output()


def _fill_suite_file(root_path, tests, suite_file, corrupt=False):  # pylint: disable=redefined-outer-name
    for test in tests:
        suite_file.write(test.get_full_address(root_path))
        if corrupt:
            suite_file.write('CORRUPT')
        suite_file.write('\n')


@pytest.fixture
def filename(tmpdir):
    return str(tmpdir.join('filename.txt'))


@pytest.fixture
def paths(request, tmpdir, use_relpath_for_dir): # pylint: disable=redefined-outer-name
    basenames = ['file{}.py'.format(i) for i in range(10)]
    basenames.extend(['file100.py:SomeClass',
                      'file101.py:Someclass.test_method'])
    returned = [os.path.join(os.path.abspath(str(tmpdir)), b) for b in basenames]

    if use_relpath_for_dir:
        dirname = str(tmpdir.join('dirname'))
        os.makedirs(dirname)
    else:
        dirname = mkdtemp()

        @request.addfinalizer
        def cleanup():  # pylint: disable=unused-variable
            shutil.rmtree(dirname)
    returned.append(dirname)

    return returned


@pytest.fixture(params=[True, False])
def use_relpath_for_dir(request):
    return request.param



from slash.exceptions import InvalidTest

def test_accidental_generator(suite, suite_test):
    suite_test.append_line('yield')
    suite_test.expect_error()
    res = suite.run()[suite_test]
    [err] = res.get_errors()
    assert err.exception_type is InvalidTest
    assert 'is a generator' in err.message



from docutils import nodes
from docutils.parsers.rst import directives, Directive
from slash import hooks  # pylint: disable=unused-import
import gossip


class HookListDoc(Directive):
    has_content = False
    required_arguments = 0
    optional_arguments = 0
    def run(self):
        returned = []
        for hook in sorted(gossip.get_group("slash").get_hooks(), key=lambda hook: hook.name):
            section = nodes.section(ids=[hook.name], names=['hooks.{}'.format(hook.name)])
            self.state.document.note_explicit_target(section)
            returned.append(section)
            title = "slash.hooks.{}".format(hook.name)
            args = hook.get_argument_names()
            if args:
                title += "({})".format(", ".join(args))
            section.append(nodes.title(text=title))
            section.append(nodes.paragraph(text=hook.doc))
        return returned

def setup(app):  # pylint: disable=unused-argument
    directives.register_directive('hook_list_doc', HookListDoc)



# pylint: disable=unused-argument,unused-variable,redefined-outer-name
import gossip
import pytest

import slash
import slash.hooks

from .conftest import Checkpoint


def test_interruption(interrupted_suite, interrupted_index):
    interrupted_suite.run(expect_interruption=True)


def test_interruption_added_to_result(interrupted_suite, interrupted_index):
    caught = []
    @gossip.register('slash.interruption_added')
    def interruption_added(result, exception):
        caught.append(exception)

    summary = interrupted_suite.run(expect_interruption=True)
    assert len(caught) == 1
    [err] = caught              # pylint: disable=unbalanced-tuple-unpacking
    assert err.exception_type is KeyboardInterrupt


def test_interruption_triggers_gossip(request, interrupted_suite, interrupted_test):
    test_id = {'value': None}

    @gossip.register('slash.test_interrupt')
    def skip():
        test_id['value'] = slash.test.__slash__.id

    @request.addfinalizer
    def cleanup():
        skip.gossip.unregister()

    summary = interrupted_suite.run(expect_interruption=True)
    assert test_id['value'] is not None
    for result in summary.get_all_results_for_test(interrupted_test):
        assert result.test_metadata.id == test_id['value']


def test_critical_cleanups_called(interrupted_suite, interrupted_test):
    cleanup = interrupted_test.add_deferred_event(
        'slash.add_critical_cleanup', 'critical_cleanup')
    summary = interrupted_suite.run(expect_interruption=True)
    assert cleanup in summary.events


def test_non_critical_cleanups_not_called(interrupted_suite, interrupted_test):
    cleanup = interrupted_test.add_cleanup()
    summary = interrupted_suite.run(expect_interruption=True)
    assert cleanup not in summary.events


def test_sigterm_interrupt(suite, suite_test):
    suite_test.append_line('raise slash.exceptions.TerminatedException()')
    suite_test.expect_interruption()
    for test in suite.iter_all_after(suite_test):
        test.expect_deselect()
    suite.run(expect_interruption=True)


@pytest.mark.parametrize('hook_name', ['session_start', 'test_start'])
def test_sigterm_on_hook(suite, hook_name):
    @gossip.register('slash.{}'.format(hook_name))
    def session_start():  # pylint: disable=unused-variable
        raise slash.exceptions.TerminatedException('Terminated by signal')

    assert suite
    for index, test in enumerate(suite):
        if index == 0 and hook_name == 'test_start':
            # first test should be interrupted...
            test.expect_interruption()
        else:
            test.expect_deselect()

    result = suite.run(expect_interruption=True)


def test_test_end_called_for_interrupted_test(interrupted_suite, interrupted_test):
    ended = []

    @gossip.register('slash.test_end')
    def test_end():
        ended.append(slash.context.test.__slash__.id)

    s = interrupted_suite.run(expect_interruption=True)
    result = s[interrupted_test]

    assert result.test_metadata.id in ended


def test_session_interruption_in_start(suite, suite_test, session_interrupt):

    @suite.slashconf.append_body
    def __code__():
        @slash.hooks.session_start.register # pylint: disable=no-member
        def session_cleanup():
            raise KeyboardInterrupt()

    for test in suite:
        test.expect_deselect()

    suite.run(expect_interruption=True)

    assert session_interrupt.called_count == 1


def test_interrupt_hooks_should_be_called_once(suite, suite_test, is_last_test, session_interrupt, test_interrupt_callback):

    @suite_test.append_body
    def __code__():
        @slash.add_critical_cleanup
        def cleanup():
            raise KeyboardInterrupt('A')
        raise KeyboardInterrupt('B')

    suite_test.expect_interruption()

    for t in suite.iter_all_after(suite_test, assert_has_more=not is_last_test):
        t.expect_deselect()

    result = suite.run(expect_interruption=True)

    assert test_interrupt_callback.called_count == 1
    assert session_interrupt.called_count == 1
    assert result.session.results.global_result.is_interrupted()


def test_interrupted_with_custom_exception(suite, suite_test, request):

    import test

    class CustomException(Exception):
        pass
    test.__interruption_exception__ = CustomException

    prev_interruption_exceptions = slash.exceptions.INTERRUPTION_EXCEPTIONS
    slash.exceptions.INTERRUPTION_EXCEPTIONS += (CustomException,)

    @request.addfinalizer
    def cleanup():
        del test.__interruption_exception__
        slash.exceptions.INTERRUPTION_EXCEPTIONS = prev_interruption_exceptions


    suite_test.append_line('import test')
    suite_test.append_line('raise test.__interruption_exception__()')
    suite_test.expect_interruption()

    for t in suite.iter_all_after(suite_test):
        t.expect_deselect()

    results = suite.run(expect_interruption=True)


def test_test_interrupt_hook_exception(suite_builder):
    # pylint: disable=reimported,redefined-outer-name
    @suite_builder.first_file.add_code
    def __code__():
        import slash

        @slash.hooks.test_interrupt.register # pylint: disable=no-member
        def test_interrupt(**_):
            1/0 # pylint: disable=pointless-statement


        def test_1():
            raise KeyboardInterrupt()

        def test_2():
            pass

    [res] = suite_builder.build().run().assert_results(1)
    assert res.is_interrupted()



@pytest.fixture
def session_interrupt():
    callback = Checkpoint()
    slash.hooks.session_interrupt.register(callback) # pylint: disable=no-member
    return callback


@pytest.fixture
def test_interrupt_callback():
    callback = Checkpoint()
    slash.hooks.test_interrupt.register(callback) # pylint: disable=no-member
    return callback


@pytest.fixture
def interrupted_suite(suite, interrupted_index):
    for index, test in enumerate(suite):
        if index == interrupted_index:
            test.append_line('raise KeyboardInterrupt()')
            test.expect_interruption()
        elif index > interrupted_index:
            test.expect_deselect()

    return suite


@pytest.fixture
def interrupted_test(interrupted_suite, interrupted_index):
    return interrupted_suite[interrupted_index]


@pytest.fixture
def interrupted_index(suite):
    return int(len(suite) // 2)



# pylint: disable=redefined-outer-name
from munch import Munch
import pytest
import gossip
import slash
from slash.utils import interactive
from slash import start_interactive_shell


def test_interactive_test(suite, interactive_checkpoint):

    summary = suite.run(additional_args=['-i'])
    assert interactive_checkpoint.called

    result = next(summary.session.results.iter_test_results())
    assert repr(result.test_metadata) == '<Interactive>'
    assert result.test_metadata.file_path == '<Interactive>'


@pytest.mark.parametrize('with_session', [True, False])
def test_interactive_scope(forge, with_session):
    interact_mock = forge.replace(interactive, '_interact')

    def _add_namespace_var(namespace):
        namespace['z'] = 4

    interact_mock({'x': 2, 'y': 3, 'z': 4})
    forge.replay()

    with gossip.registered('slash.before_interactive_shell', _add_namespace_var):
        if with_session:
            with slash.Session():
                slash.g.x = 2
                start_interactive_shell(y=3)
        else:
            start_interactive_shell(x=2, y=3)


def test_interactive_planned_tests(interactive_checkpoint, suite):

    counts = Munch(original=len(suite), got=None)

    @slash.hooks.register
    def tests_loaded(tests):    # pylint: disable=unused-variable
        counts.got = len(tests)

    suite.run(additional_args=['-i'])
    assert interactive_checkpoint.called
    assert counts.got == counts.original + 1



@pytest.fixture
def interactive_checkpoint(checkpoint, forge):
    def _interact(*_, **__):  # pylint: disable=unused-argument
        assert slash.context.session.scope_manager.get_current_stack() == ['session', 'module', 'test']
        assert slash.context.test.__slash__.is_interactive()
        assert slash.context.test.__slash__.id
        checkpoint()

    forge.replace_with(interactive, '_interact', _interact)

    return checkpoint



import pytest


@pytest.mark.parametrize('parametrize', [True, False])
def test_repeating_each_test(suite, suite_test, parametrize):
    num_repetitions = 5
    suite_test.repeat(num_repetitions)

    if parametrize:
        param = suite_test.add_parameter()

    summary = suite.run()
    results = summary.get_all_results_for_test(suite_test)
    if parametrize:
        assert len(results) == num_repetitions * len(param.values)
    else:
        assert len(results) == num_repetitions
    assert len(set(result.test_metadata.id for result in results)) == len(results)


def test_repeat_each_global(suite, config_override):
    num_repetitions = 5
    config_override('run.repeat_each', num_repetitions)
    for test in suite:
        test.expect_repetition(num_repetitions)
    assert suite.run().ok()


def test_repeat_all_global(suite, config_override):
    num_repetitions = 5
    config_override('run.repeat_all', num_repetitions)
    for index, test in enumerate(suite):
        test.append_line('slash.context.result.data["index"]={}'.format(index))

    for test in suite:
        test.expect_repetition(num_repetitions)

    summary = suite.run()

    indices = [res.data['index'] for res in summary.session.results.iter_test_results()]

    expected = [[x for x in range(len(suite))] for _ in range(num_repetitions)]
    assert indices == [item for sublist in expected for item in sublist]


@pytest.mark.parametrize('repeat_mode', ['repeat_each', 'repeat_all'])
def test_repeat_isolation(suite, suite_test, config_override, repeat_mode):
    num_repetitions = 5
    config_override('run.{}'.format(repeat_mode), num_repetitions)
    suite_test.append_line('import time')
    suite_test.append_line('slash.context.result.data["time"] = time.time()')

    for test in suite:
        test.expect_repetition(num_repetitions)

    summary = suite.run()
    results = summary.get_all_results_for_test(suite_test)
    assert len(results) == num_repetitions
    assert len({result.data['time'] for result in results}) == len(results)



from slash.utils.exception_mark import ExceptionMarker, mark_exception, get_exception_mark

def test_marker():
    exc = AttributeError()
    marker = ExceptionMarker('mark')
    assert not marker.is_exception_marked(exc)
    marker.mark_exception(exc)
    assert marker.is_exception_marked(exc)


def test_marker_mark_returns_exception():
    exc = AttributeError()
    assert ExceptionMarker('mark').mark_exception(exc) is exc
    assert ExceptionMarker('mark').is_exception_marked(exc)

def test_marker_class():

    class CustomException(Exception):
        pass
    mark_exception(CustomException, 'a', 'b')
    assert get_exception_mark(CustomException, 'a') == 'b'

    e = CustomException()
    mark_exception(e, 'c', 'd')
    assert get_exception_mark(e, 'a') == 'b'
    assert get_exception_mark(e, 'c') == 'd'
    assert get_exception_mark(CustomException, 'c') is None



from slash.app import Application

from slash.reporting.console_reporter import ConsoleReporter


def test_app_reporter():
    with Application() as app:
        assert isinstance(app.session.reporter, ConsoleReporter)

def test_custom_reporter():

    class DummyReporter(object):
        pass

    dummy_reporter = DummyReporter()

    app = Application()
    app.set_reporter(dummy_reporter)
    with app:
        assert app.session.reporter is dummy_reporter



from __future__ import print_function
import argparse
import os
import sys
from contextlib import contextmanager

import colorama

from .. import conf, plugins
from .._compat import itervalues


_PLUGIN_ACTIVATION_PREFIX = "--with-"
_PLUGIN_DEACTIVATION_PREFIX = "--without-"

def add_pending_plugins_from_commandline(argv):
    returned_argv = []
    for arg in argv:
        if arg.startswith(_PLUGIN_DEACTIVATION_PREFIX):
            plugin_name = arg[len(_PLUGIN_DEACTIVATION_PREFIX):]
            plugins.manager.deactivate_later(plugin_name)
        elif arg.startswith(_PLUGIN_ACTIVATION_PREFIX):
            plugin_name = arg[len(_PLUGIN_ACTIVATION_PREFIX):]
            plugins.manager.activate_later(plugin_name)
        else:
            returned_argv.append(arg)
    return returned_argv

def configure_arg_parser_by_plugins(parser):
    for plugin in itervalues(plugins.manager.get_installed_plugins()):
        group = parser.add_argument_group('Options for {}{}'.format(
            '' if plugins.manager.is_internal_plugin(plugin) else '--with-',
            plugins.manager.normalize_command_line_name(plugin.get_name())))
        plugin.configure_argument_parser(group)

def configure_arg_parser_by_config(parser, config=None):
    if config is None:
        config = conf.config

    plugin_groups = {}

    parser.add_argument(
        "-o", dest="config_overrides", metavar="PATH=VALUE", action="append",
        default=[],
        help="Provide overrides for configuration"
    )
    for path, node, cmdline in _iter_cmdline_config(config):
        if path.startswith('plugin_config.'):
            plugin_name = path.split('.')[1]
            subparser = plugin_groups.get(plugin_name)
            if subparser is None:
                subparser = plugin_groups[plugin_name] = parser.add_argument_group(
                    'options for the {0} plugin (--with-{0})'.format(plugin_name))
        else:
            subparser = parser
        cmdline.configure_parser(subparser, path, node)

def configure_plugins_from_args(args):
    for plugin in itervalues(plugins.manager.get_active_plugins()):
        plugin.configure_from_parsed_args(args)

def _iter_cmdline_config(config):
    for path, cfg in config.traverse_leaves():
        cmdline = (cfg.metadata or {}).get("cmdline")
        if cmdline is None:
            continue
        yield path, cfg, cmdline

@contextmanager
def get_modified_configuration_from_args_context(parser, args, config=None):
    if config is None:
        config = conf.config
    to_restore = []
    try:
        for path, cfg, cmdline in _iter_cmdline_config(config):
            old_value = cfg.get_value()
            new_value = cmdline.update_value(old_value, args)
            if new_value != old_value:
                to_restore.append((path, cfg.get_value()))
                try:
                    config.assign_path(path, new_value, deduce_type=True, default_type=str)
                except ValueError:
                    parser.error('Invalid value for {}: {!r}'.format(cmdline, new_value))
        for override in args.config_overrides:
            if "=" not in override:
                parser.error("Invalid config override: {}".format(override))
            path, _ = override.split("=", 1)
            to_restore.append((path, config.get_path(path)))
            try:
                config.assign_path_expression(override, deduce_type=True, default_type=str)
            except ValueError:
                parser.error("Invalid value for config override: {}".format(override))
        yield
    finally:
        for path, prev_value in reversed(to_restore):
            config.assign_path(path, prev_value)

class SlashArgumentParser(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        super(SlashArgumentParser, self).__init__(
            prog=self._deduce_program_name(),
            usage='{} [options]'.format(self._deduce_program_name()),
            *args, **kwargs)

        self._positionals_metavar = None

    def _deduce_program_name(self):
        returned = os.path.basename(sys.argv[0])
        if len(sys.argv) > 1:
            returned += " {}".format(sys.argv[1])
        return returned

    def set_positional_metavar(self, metavar, plural=True):
        self._positionals_metavar = metavar
        if plural:
            self.usage += ' {0} [{0} [...]]'.format(metavar)
        else:
            self.usage += ' {}'.format(metavar)

    def set_description(self, description):
        self.description = description


COLOR_RESET = colorama.Fore.RESET + colorama.Back.RESET + colorama.Style.RESET_ALL  # pylint: disable=no-member


class ColorizedString(str):

    def __new__(cls, *args, **kwargs):
        style = kwargs.pop('style', None)
        returned = str.__new__(cls, *args, **kwargs)
        returned.style = style
        return returned

    def colorize(self):
        if self.style:
            return '{}{}{}'.format(self.style, self, COLOR_RESET)
        return str(self)

def make_styler(style):
    return lambda string: ColorizedString(string, style=style)

UNDERLINED = '\x1b[4m'


class Printer(object):

    def __init__(self, report_stread, enable_output=True, force_color=False, enable_color=True):
        self._stream = report_stread
        self._output_enabled = enable_output
        self._force_color = force_color
        self._color_enabled = enable_color

    def _colored_print(self, *args, **print_kwargs):
        self._print(*(getattr(arg, 'colorize', arg.__str__)() for arg in args), **print_kwargs)

    def _print(self, *args, **print_kwargs):
        print(*args, file=self._stream, **print_kwargs)

    def __call__(self, *args, **print_kwargs):
        if self._output_enabled:
            should_color = self._force_color or (self._color_enabled and self._stream.isatty())
            print_func = self._colored_print if should_color else self._print
            print_func(*args, **print_kwargs)


def error_abort(message, *args, **kwargs):
    stream = kwargs.pop('stream', sys.stderr)
    assert not kwargs
    if args:
        message = message.format(*args)
    print(message, file=stream)
    sys.exit(-1)



# pylint: disable=unused-argument, unused-variable, redefined-outer-name
import slash
import gossip
import pytest

from .utils import NamedPlugin, CustomException, make_runnable_tests


@pytest.mark.parametrize('has_start', [True, False])
def test_plugin_end_not_called_when_start_not_called(gossip_raise_immediately, has_start, checkpoint, restore_plugins_on_cleanup):
    errors = []

    class CustomPlugin(NamedPlugin):
        pass

    class RaisingPlugin(NamedPlugin):
        pass

    if has_start:
        def start(self):
            errors.append('start hook called')
        setattr(CustomPlugin, 'session_start', start)

    def end(self):
        errors.append('end hook called')
    setattr(CustomPlugin, 'session_end', end)

    def do_raise(self):
        checkpoint()
        raise CustomException()
    setattr(RaisingPlugin, 'session_start', do_raise)


    slash.plugins.manager.install(RaisingPlugin(), activate=True)
    slash.plugins.manager.install(CustomPlugin(), activate=True)


    with slash.Session() as s:
        with pytest.raises(CustomException):
            with s.get_started_context():
                def test_something():
                    pass
                slash.runner.run_tests(make_runnable_tests(test_something))

    assert checkpoint.called
    assert not errors


@pytest.mark.parametrize('illegal_char', ['\t', '\n', '$', '(', '[', '{', '='])
def test_plugin_with_illegal_name(illegal_char):

    class PluginWithIllegalName(slash.plugins.PluginInterface):
        def __init__(self):
            self._name = 'ab' + illegal_char

        def get_name(self):
            return self._name

    plugin = PluginWithIllegalName()
    with pytest.raises(slash.plugins.IllegalPluginName):
        slash.plugins.manager.install(plugin)


def test_plugin_with_no_session_start_gets_called_session_end(checkpoint):

    @slash.plugins.active
    class MyPlugin(NamedPlugin):

        def session_end(self):
            checkpoint()

    with slash.Session() as s:
        with s.get_started_context():
            pass

    assert checkpoint.called


@pytest.yield_fixture
def gossip_raise_immediately():
    g = gossip.get_group('slash')
    prev = g.get_exception_policy()
    g.set_exception_policy(gossip.RaiseImmediately())
    try:
        yield
    finally:
        g.set_exception_policy(prev)



import itertools
import traceback
import os
import sys
from types import FunctionType, GeneratorType
from contextlib import contextmanager

import dessert
from emport import import_file
from logbook import Logger
from sentinels import NOTHING


from .conf import config
from ._compat import string_types
from .ctx import context
from .core.local_config import LocalConfig
from .core.markers import repeat_marker
from . import hooks
from .core.runnable_test import RunnableTest
from .core.test import Test, TestTestFactory, is_valid_test_name
from .core.function_test import FunctionTestFactory
from .exception_handling import handling_exceptions, mark_exception_handled, get_exception_frame_correction
from .exceptions import CannotLoadTests
from .core.runnable_test_factory import RunnableTestFactory
from .utils.pattern_matching import Matcher
from .utils.python import check_duplicate_functions
from .resuming import ResumedTestData
from .utils.interactive import generate_interactive_test

_logger = Logger(__name__)


class Loader(object):

    """
    Provides iteration interfaces to load runnable tests from various places
    """

    def __init__(self):
        super(Loader, self).__init__()
        self._local_config = LocalConfig()
        self._duplicate_funcs = set()

    _cached_matchers = NOTHING

    def _get_matchers(self):
        if self._cached_matchers is NOTHING:
            if config.root.run.filter_strings:
                self._cached_matchers = [Matcher(s) for s in config.root.run.filter_strings]
            else:
                self._cached_matchers = None
        return self._cached_matchers


    def get_runnables(self, paths, prepend_interactive=False):
        assert context.session is not None
        sources = self._generate_repeats(self._generate_test_sources(paths))
        returned = self._collect(sources)
        self._duplicate_funcs |= self._local_config.duplicate_funcs
        for (path, name, line) in sorted(self._duplicate_funcs):
            _logger.warning('Duplicate function definition, File: {}, Name: {}, Line: {}', path, name, line)

        if prepend_interactive:
            returned.insert(0, generate_interactive_test())

        hooks.tests_loaded(tests=returned) # pylint: disable=no-member
        returned.sort(key=lambda test: (
            test.__slash__.repeat_all_index, test.__slash__.get_sort_key()
        ))
        return returned


    def _generate_repeats(self, tests):
        returned = []
        repeat_each = config.root.run.repeat_each
        for test in tests:
            for i in range(repeat_each * repeat_marker.get_value(test.get_test_function(), 1)):
                returned.append(test.clone() if i else test)
        num_tests = len(returned)
        for i in range(config.root.run.repeat_all - 1):
            for test in itertools.islice(returned, 0, num_tests):
                clone = test.clone()
                clone.__slash__.repeat_all_index = i + 1
                returned.append(clone)
        return returned


    def _collect(self, iterator):
        returned = []
        context.reporter.report_collection_start()
        try:
            for x in iterator:
                assert x.__slash__.id is None
                x.__slash__.allocate_id()
                returned.append(x)
                context.reporter.report_test_collected(returned, x)
        finally:
            context.reporter.report_collection_end(returned)
        context.session.increment_total_num_tests(len(returned))
        return returned

    def _generate_test_sources(self, thing, matcher=None):

        if isinstance(thing, tuple):
            assert len(thing) == 2, '_generate_test_sources on tuples requires a tuple of (loadable_obj, matcher)'
            iterator = self._generate_test_sources(thing[0], matcher=thing[1])

        elif isinstance(thing, (list, GeneratorType, itertools.chain)):
            iterator = itertools.chain.from_iterable(self._generate_test_sources(x) for x in thing)
        elif isinstance(thing, string_types):
            iterator = self._iter_test_address(thing)
        elif isinstance(thing, RunnableTest):
            iterator = [thing]
        elif isinstance(thing, ResumedTestData):
            iterator = self._iter_test_resume(thing)
        elif not isinstance(thing, RunnableTestFactory):
            thing = self._get_runnable_test_factory(thing)
            iterator = thing.generate_tests(fixture_store=context.session.fixture_store)

        return (t for t in iterator if matcher is None or matcher.matches(t.__slash__))

    def _iter_test_resume(self, resume_state):
        for test in self._iter_path(resume_state.file_name):
            if resume_state.function_name == test.__slash__.address_in_file:
                if resume_state.variation:
                    if not resume_state.variation == test.get_variation().id:
                        continue
                yield test

    def _iter_test_address(self, address):
        drive, address = os.path.splitdrive(address)
        if ':' in address:
            path, address_in_file = address.split(':', 1)
        else:
            path = address
            address_in_file = None
        path = os.path.join(drive, path)


        tests = list(self._iter_path(path))

        # special case for directories where we couldn't load any tests (without filter)
        if not tests and address_in_file is None:
            return

        matched = False
        for test in tests:

            if address_in_file is not None:
                if not self._address_in_file_matches(address_in_file, test):
                    continue
            matched = True
            yield test
        if not matched:
            raise CannotLoadTests('Cannot find test(s) for {!r}'.format(address))

    def _address_in_file_matches(self, address_in_file, test):
        if address_in_file == test.__slash__.factory_name:
            return True
        test_address_in_file = test.__slash__.address_in_file
        if address_in_file == test_address_in_file:
            return True
        if '(' in test_address_in_file:
            if address_in_file == test_address_in_file[:test_address_in_file.index('(')]:
                return True
        return False

    def _iter_path(self, path):
        return self._iter_paths([path])

    def _iter_paths(self, paths):

        paths = list(paths)
        for path in paths:
            if not os.path.exists(path):
                msg = "Path {!r} could not be found".format(path)
                with handling_exceptions():
                    raise CannotLoadTests(msg)
        for path in paths:
            for file_path in _walk(path):
                _logger.debug("Checking {0}", file_path)
                if not self._is_file_wanted(file_path):
                    _logger.debug("{0} is not wanted. Skipping...", file_path)
                    continue
                module = None
                try:
                    with handling_exceptions(context="during import"):
                        if not config.root.run.message_assertion_introspection:
                            dessert.disable_message_introspection()
                        with dessert.rewrite_assertions_context():
                            module = import_file(file_path)
                except Exception as e:

                    tb_file, tb_lineno, _, _ = _extract_tb()
                    raise mark_exception_handled(
                        CannotLoadTests(
                            "Could not load {0!r} ({1}:{2} - {3})".format(file_path, tb_file, tb_lineno, e)))
                if module is not None:
                    self._duplicate_funcs |= check_duplicate_functions(file_path)
                    with self._adding_local_fixtures(file_path, module):
                        for runnable in self._iter_runnable_tests_in_module(file_path, module):
                            if self._is_excluded(runnable):
                                continue
                            yield runnable

    @contextmanager
    def _adding_local_fixtures(self, file_path, module):
        with context.session.fixture_store.new_namespace_context():
            self._local_config.push_path(os.path.dirname(file_path))
            try:
                context.session.fixture_store.add_fixtures_from_dict(
                    self._local_config.get_dict())
                with context.session.fixture_store.new_namespace_context():
                    context.session.fixture_store.add_fixtures_from_dict(
                        vars(module))
                    context.session.fixture_store.resolve()
                    yield
            finally:
                self._local_config.pop_path()


    def _is_excluded(self, test):
        matchers = self._get_matchers()
        if matchers is None:
            return False
        return not all(m.matches(test.__slash__) for m in matchers)

    def _is_file_wanted(self, filename):
        return filename.endswith(".py")

    def _iter_runnable_tests_in_module(self, file_path, module):
        for thing_name in sorted(dir(module)):
            thing = getattr(module, thing_name)
            if thing is RunnableTestFactory:  # probably imported directly
                continue

            factory = self._get_runnable_test_factory(thing)

            if factory is None:
                continue

            factory.set_factory_name(thing_name)
            factory.set_module_name(module.__name__)
            factory.set_filename(file_path)

            for test in factory.generate_tests(fixture_store=context.session.fixture_store):
                assert test.__slash__ is not None
                yield test

    def _get_runnable_test_factory(self, thing):

        if isinstance(thing, type) and issubclass(thing, Test):
            return TestTestFactory(thing)

        if isinstance(thing, FunctionType):
            if is_valid_test_name(thing.__name__):
                return FunctionTestFactory(thing)

        return None


def _walk(p):
    if os.path.isfile(p):
        yield p
        return

    for path, dirnames, filenames in os.walk(p):
        dirnames[:] = sorted(dirname for dirname in dirnames if not dirname.startswith('.'))
        for filename in sorted(filenames):
            yield os.path.join(path, filename)

def _extract_tb():
    _, exc_value, exc_tb = sys.exc_info()
    returned = traceback.extract_tb(exc_tb)
    return returned[-1 - get_exception_frame_correction(exc_value)]



import slash
import os
import signal
from .utils.suite_writer import Suite
from slash.resuming import get_tests_from_previous_session
from slash.exceptions import InteractiveParallelNotAllowed, ParallelTimeout
from slash.parallel.server import ServerStates
from slash.parallel.parallel_manager import ParallelManager
from slash import Session
from slash.loader import Loader
import time
import tempfile
import pytest

@pytest.fixture(scope='module', autouse=True)
def no_parallel_user_config(request):
    tmpdir = tempfile.mkdtemp()
    path = os.path.join(tmpdir, 'slashrc')
    os.environ["SLASH_USER_SETTINGS"] = path

    @request.addfinalizer
    def cleanup():  # pylint: disable=unused-variable
        os.rmdir(tmpdir)
        del os.environ["SLASH_USER_SETTINGS"]

#basic features of parallel
def run_specific_workers_and_tests_num(workers_num, tests_num=10):
    suite = Suite(debug_info=False, is_parallel=True)
    suite.populate(num_tests=tests_num)
    summary = suite.run(num_workers=workers_num)
    assert len(summary.session.parallel_manager.server.worker_session_ids) == workers_num
    assert summary.session.results.get_num_successful() == tests_num
    assert summary.session.results.is_success()
    return summary

def test_one_worker():
    run_specific_workers_and_tests_num(workers_num=1)

def test_many_workers():
    run_specific_workers_and_tests_num(workers_num=3, tests_num=50)

def test_zero_workers(parallel_suite):
    summary = parallel_suite.run(num_workers=0) #should act like regular run of slash, not parallel
    assert summary.session.results.is_success()
    assert summary.session.parallel_manager is None

def test_test_causes_worker_exit(parallel_suite, config_override):
    config_override("parallel.communication_timeout_secs", 2)
    parallel_suite[0].append_line("import os")
    parallel_suite[0].append_line("os._exit(0)")
    parallel_suite[0].expect_interruption()
    workers_num = 1
    summary = parallel_suite.run(num_workers=workers_num, verify=False)
    assert len(summary.session.parallel_manager.server.worker_session_ids) == workers_num
    [result] = summary.get_all_results_for_test(parallel_suite[0])
    assert result.is_interrupted()

def test_keepalive_works(parallel_suite, config_override):
    config_override("parallel.communication_timeout_secs", 2)
    parallel_suite[0].append_line("import time")
    parallel_suite[0].append_line("time.sleep(6)")
    workers_num = 1
    summary = parallel_suite.run(num_workers=workers_num)
    assert len(summary.session.parallel_manager.server.worker_session_ids) == workers_num
    assert summary.session.results.is_success()

def test_server_fails(parallel_suite):

    @slash.hooks.worker_connected.register  # pylint: disable=no-member, unused-argument
    def simulate_ctrl_c(session_id):  # pylint: disable=unused-variable, unused-argument
        pid = os.getpid()
        os.kill(pid, signal.SIGINT)

    @slash.hooks.session_interrupt.register  # pylint: disable=no-member
    def check_workers_and_server_down():  # pylint: disable=unused-variable
        for worker in slash.context.session.parallel_manager.workers.values():
            ret = worker.poll()
            assert not ret is None
        assert slash.context.session.parallel_manager.server.interrupted
        assert not slash.context.session.parallel_manager.server.finished_tests

    for test in parallel_suite:
        test.expect_deselect()
    parallel_suite.run(expect_interruption=True)


#test slash features with parallel
def test_test_success(parallel_suite):
    results = parallel_suite.run().session.results
    assert results.is_success()
    assert results.get_num_successful() == len(parallel_suite)

def test_test_failure(parallel_suite):
    parallel_suite[0].when_run.fail()
    summary = parallel_suite.run()
    [result] = summary.get_all_results_for_test(parallel_suite[0])
    [failures] = result.get_failures()
    assert 'AssertionError' in str(failures)
    assert 'assert False' in str(failures)

def test_stop_on_error(parallel_suite, parallel_suite_test):
    parallel_suite_test.when_run.fail()
    summary = parallel_suite.run(additional_args=['-x'], verify=False)
    [result] = summary.get_all_results_for_test(parallel_suite_test)
    assert result.is_failure()

    found_failure = False
    for result in summary.session.results:
        if result.is_failure():
            found_failure = True
            continue
        if found_failure:
            assert result.is_not_run()
    assert found_failure

def test_pass_override_conf_flag(parallel_suite):
    summary = parallel_suite.run(additional_args=['-o', 'parallel.server_port=8001'])
    results = summary.session.results
    assert results.is_success()
    assert results.get_num_successful() == len(parallel_suite)
    assert summary.session.parallel_manager.server.port == 8001

def test_test_error(parallel_suite):
    parallel_suite[0].append_line('slash.add_error()')
    parallel_suite[0].expect_error()
    summary = parallel_suite.run()
    [result] = summary.get_all_results_for_test(parallel_suite[0])
    [err] = result.get_errors()
    assert 'RuntimeError' in str(err)
    assert 'add_error() must be called' in str(err)

def test_test_interruption_causes_communication_timeout(parallel_suite, config_override):
    config_override("parallel.communication_timeout_secs", 2)
    parallel_suite[0].when_run.interrupt()
    summary = parallel_suite.run(num_workers=1, verify=False)
    [interrupted_result] = summary.get_all_results_for_test(parallel_suite[0])
    assert interrupted_result.is_interrupted()
    for result in summary.session.results:
        if result != interrupted_result:
            assert result.is_success() or result.is_not_run()

def test_test_interruption_causes_no_requests(parallel_suite, config_override):
    config_override("parallel.no_request_timeout", 2)
    parallel_suite[0].when_run.interrupt()
    summary = parallel_suite.run(num_workers=1, verify=False)
    assert summary.get_all_results_for_test(parallel_suite[0]) == []

def test_test_skips(parallel_suite):
    parallel_suite[0].add_decorator('slash.skipped("reason")')
    parallel_suite[0].expect_skip()
    results = parallel_suite.run().session.results
    assert results.is_success(allow_skips=True)
    assert results.get_num_skipped() == 1
    for result in results:
        if result.is_skip():
            assert 'reason' in result.get_skips()

def test_session_warnings(parallel_suite):
    parallel_suite[0].append_line("import warnings")
    parallel_suite[0].append_line("warnings.warn('message')")
    session_results = parallel_suite.run().session
    assert len(session_results.warnings) == 1
    [w] = session_results.warnings
    assert w.message == 'message'

def test_child_session_errors(parallel_suite):
    parallel_suite[0].expect_failure()
    parallel_suite[0].append_line("import slash")
    parallel_suite[0].append_line("slash.context.session.results.global_result.add_error('bla')")
    session_results = parallel_suite.run(num_workers=1, verify=False).session
    assert not session_results.results.is_success()
    assert session_results.parallel_manager.server.worker_session_error_reported

def test_child_errors_in_cleanup_are_session_errors(parallel_suite):
    parallel_suite[0].expect_failure()
    parallel_suite[0].append_line("import slash")
    parallel_suite[0].append_line("def a():")
    parallel_suite[0].append_line("   def _cleanup():")
    parallel_suite[0].append_line("      slash.add_error('Session cleanup')")
    parallel_suite[0].append_line("   slash.add_cleanup(_cleanup, scope='session')")
    parallel_suite[0].append_line("a()")
    session_results = parallel_suite.run(num_workers=1, verify=False).session
    assert not session_results.results.is_success()
    assert session_results.parallel_manager.server.worker_session_error_reported

def test_traceback_vars(parallel_suite):
    #code to be inserted:
        #     def test_traceback_frames():
        #     num = 0
        #     a()
        #
        # def a():
        #     x=1
        #     assert False
    parallel_suite[0].append_line("def a():")
    parallel_suite[0].append_line("   num = 0")
    parallel_suite[0].append_line("   b()")
    parallel_suite[0].append_line("def b():")
    parallel_suite[0].append_line("   x=1")
    parallel_suite[0].append_line("   assert False")
    parallel_suite[0].append_line("a()")
    parallel_suite[0].expect_failure()
    summary = parallel_suite.run(num_workers=1)
    results = summary.session.results
    found_failure = 0
    for result in results:
        if result.is_failure():
            found_failure += 1
            assert len(result.get_failures()) == 1
            assert len(result.get_failures()[0].traceback.frames) == 3
            assert 'x' in result.get_failures()[0].traceback.frames[2].locals
            assert 'num' in result.get_failures()[0].traceback.frames[1].locals
    assert found_failure == 1

def test_result_data_not_picklable(parallel_suite):
    parallel_suite[0].append_line("import socket")
    parallel_suite[0].append_line("s = socket.socket()")
    parallel_suite[0].append_line("slash.context.result.data.setdefault('socket', s)")
    summary = parallel_suite.run()
    [result] = summary.get_all_results_for_test(parallel_suite[0])
    assert result.data == {}

def test_result_data_is_picklable(parallel_suite):
    parallel_suite[0].append_line("slash.context.result.data.setdefault('num', 1)")
    summary = parallel_suite.run()
    [result] = summary.get_all_results_for_test(parallel_suite[0])
    assert 'num' in result.data
    assert result.data['num'] == 1

def test_result_details_not_picklable(parallel_suite):
    parallel_suite[0].append_line("import socket")
    parallel_suite[0].append_line("s = socket.socket()")
    parallel_suite[0].append_line("slash.context.result.details.append('socket', s)")
    summary = parallel_suite.run()
    [result] = summary.get_all_results_for_test(parallel_suite[0])
    assert result.details.all() == {}

def test_result_details_is_picklable(parallel_suite):
    parallel_suite[0].append_line("slash.context.result.details.append('num', 1)")
    summary = parallel_suite.run()
    [result] = summary.get_all_results_for_test(parallel_suite[0])
    details = result.details.all()
    assert 'num' in details
    assert details['num'] == [1]

def test_parameters(parallel_suite):
    parallel_suite[0].add_parameter(num_values=1)
    summary = parallel_suite.run()
    assert summary.session.results.is_success()

def test_requirements(parallel_suite):
    parallel_suite[0].add_decorator('slash.requires(False)')
    parallel_suite[0].expect_skip()
    results = parallel_suite.run().session.results
    assert results.get_num_skipped() == 1
    assert results.get_num_successful() == len(parallel_suite) - 1
    assert results.is_success(allow_skips=True)

def test_is_test_code(parallel_suite):
    parallel_suite[0].when_run.error()
    summary = parallel_suite.run()
    [result] = summary.get_all_results_for_test(parallel_suite[0])
    [err] = result.get_errors()
    assert err.traceback.frames[-1].is_in_test_code()
    error_json = err.traceback.to_list()
    assert error_json[-1]['is_in_test_code']

def test_parallel_resume(parallel_suite):
    parallel_suite[0].when_run.fail()
    result = parallel_suite.run()
    resumed = get_tests_from_previous_session(result.session.id)
    assert len(resumed) == 1

def test_parallel_symlinks(parallel_suite, logs_dir):   # pylint: disable=unused-argument
    files_dir = logs_dir.join("files")
    links_dir = logs_dir.join("links")
    session = parallel_suite.run(additional_args=['-l', str(files_dir)]).session
    session_log_file = files_dir.join(session.id, "session.log")

    assert session.results.is_success()
    assert session_log_file.check()
    assert links_dir.join("last-session").readlink() == session_log_file
    assert links_dir.join("last-session-dir").readlink() == session_log_file.dirname

    worker_session_ids = session.parallel_manager.server.worker_session_ids
    file_names = [x.basename for x in links_dir.join("last-session-dir").listdir()]
    assert 'worker_1' in file_names

    for file_name in links_dir.join("last-session-dir").listdir():
        if file_name.islink() and 'worker' in file_name.basename:
            last_token = file_name.readlink().split('/')[-1]
            assert last_token in worker_session_ids
            assert os.path.isdir(file_name.readlink())

def test_parallel_interactive_fails(parallel_suite):
    summary = parallel_suite.run(additional_args=['-i'], verify=False)
    results = list(summary.session.results.iter_all_results())
    assert len(results) == 1
    error = results[0].get_errors()[0]
    assert error.exception_type == InteractiveParallelNotAllowed

def test_children_session_ids(parallel_suite):
    summary = parallel_suite.run()
    assert summary.session.results.is_success()
    session_ids = summary.session.parallel_manager.server.worker_session_ids
    expected_session_ids = ["{}_1".format(summary.session.id.split('_')[0])]
    assert session_ids == expected_session_ids

def test_timeout_no_request_to_server(config_override, runnable_test_dir):
    config_override("parallel.no_request_timeout", 1)
    with Session():
        runnables = Loader().get_runnables(str(runnable_test_dir))
        parallel_manager = ParallelManager([])
        parallel_manager.start_server_in_thread(runnables)
        parallel_manager.server.state = ServerStates.SERVE_TESTS

        with slash.assert_raises(ParallelTimeout) as caught:
            parallel_manager.start()
        assert 'No request sent to server' in caught.exception.args[0]

def test_children_not_connected_timeout(runnable_test_dir, config_override):
    config_override("parallel.worker_connect_timeout", 0)
    config_override("parallel.num_workers", 1)
    with Session():
        runnables = Loader().get_runnables(str(runnable_test_dir))
        parallel_manager = ParallelManager([])
        parallel_manager.start_server_in_thread(runnables)
        time.sleep(0.1)
        with slash.assert_raises(ParallelTimeout) as caught:
            parallel_manager.wait_all_workers_to_connect()
        assert caught.exception.args[0] == 'Not all clients connected'

def test_worker_error_logs(parallel_suite, config_override):
    config_override("parallel.communication_timeout_secs", 2)
    parallel_suite[0].when_run.interrupt()
    summary = parallel_suite.run(num_workers=1, verify=False)
    [interrupted_result] = summary.get_all_results_for_test(parallel_suite[0])
    assert interrupted_result.is_interrupted()
    for result in summary.session.results:
        if result != interrupted_result:
            assert result.is_success() or result.is_not_run()
    file_path = os.path.join(summary.session.parallel_manager.workers_error_dircetory, 'errors-worker-1.log')
    assert os.path.isfile(file_path)
    with open(file_path) as error_file:
        line = error_file.readline()
        assert 'interrupted' in line



import functools
import sys
import logbook
from ..app import Application
from ..conf import config
from ..exception_handling import handling_exceptions
from ..exceptions import CannotLoadTests, InteractiveParallelNotAllowed
from ..resuming import (get_last_resumeable_session_id, get_tests_from_previous_session, save_resume_state, clean_old_entries)
from ..runner import run_tests
from ..utils.suite_files import iter_suite_file_paths
from ..utils.tmux_utils import run_slash_in_tmux
from ..plugins import manager
from ..parallel.parallel_manager import ParallelManager
from ..parallel.worker import Worker

_logger = logbook.Logger(__name__)

def slash_run(args, report_stream=None, resume=False, rerun=False, app_callback=None, working_directory=None):

    if report_stream is None:
        report_stream = sys.stderr
    app = Application()
    if resume:
        app.arg_parser.set_positional_metavar('SESSION-ID', plural=False)
        app.arg_parser.set_description('Resumes a previously run session by running only unsuccessful on unfinished tests')
    else:
        app.arg_parser.set_positional_metavar('TEST')
    if working_directory is not None:
        app.set_working_directory(working_directory)
    app.set_argv(args)
    app.set_report_stream(report_stream)
    app.enable_interactive()
    collected = []
    try:
        with app:
            if app_callback is not None:
                app_callback(app)
            try:
                with handling_exceptions():
                    if is_parallel() and app.parsed_args.interactive:
                        raise InteractiveParallelNotAllowed("Cannot run interactive mode in parallel")
                    if config.root.tmux.enabled and not is_child():
                        _logger.notice("About to start slash in new tmux session...")
                        run_slash_in_tmux(args)
                    if is_child():
                        worker = Worker(config.root.parallel.worker_id, app.session.id)
                        worker.connect_to_server()
                    if resume or rerun:
                        session_ids = app.positional_args
                        if not session_ids:
                            session_ids = [get_last_resumeable_session_id()]
                        get_successful_tests = False if resume else True
                        to_resume = [x for session_id in session_ids for x in \
                                     get_tests_from_previous_session(session_id, get_successful_tests=get_successful_tests)]
                        collected = app.test_loader.get_runnables(to_resume, prepend_interactive=app.parsed_args.interactive)
                    else:
                        collected = _collect_tests(app, args)

                collected = list(collected)
                if is_child():
                    worker.start_execution(app, collected)
                else:
                    with app.session.get_started_context():
                        report_tests_to_backslash(collected)
                        if is_parent():
                            app.session.parallel_manager = ParallelManager(args)
                            app.session.parallel_manager.start_server_in_thread(collected)
                            app.session.parallel_manager.start()
                        else:
                            run_tests(collected)

            finally:
                save_resume_state(app.session.results, collected)
                clean_old_entries()
            if app.exit_code == 0 and not app.session.results.is_success(allow_skips=True):
                app.set_exit_code(-1)
    except Exception:         # pylint: disable=broad-except
        # Error reporting happens in app context
        assert app.exit_code != 0

    return app

slash_resume = functools.partial(slash_run, resume=True)
slash_rerun = functools.partial(slash_run, rerun=True)

def is_parallel():
    return config.root.parallel.num_workers

def is_child():
    return config.root.parallel.worker_id is not None

def is_parent():
    return is_parallel() and not is_child()

def report_tests_to_backslash(tests):
    active_plugins = manager.get_active_plugins()
    backslash_plugin = active_plugins.get('backslash', None)
    if backslash_plugin and hasattr(backslash_plugin, 'report_planned_tests'):
        backslash_plugin.report_planned_tests(tests)

def _collect_tests(app, args):  # pylint: disable=unused-argument
    paths = app.positional_args

    paths = _extend_paths_from_suite_files(paths)

    if not paths and not app.parsed_args.interactive:
        paths = config.root.run.default_sources

    if not paths and not app.parsed_args.interactive:
        raise CannotLoadTests("No tests specified")

    collected = app.test_loader.get_runnables(paths, prepend_interactive=app.parsed_args.interactive)
    if not collected:
        raise CannotLoadTests("No tests could be collected")

    return collected

def _extend_paths_from_suite_files(paths):
    suite_files = config.root.run.suite_files
    if not suite_files:
        return paths
    paths = list(paths)
    paths.extend(iter_suite_file_paths(suite_files))
    return paths



import slash
from .utils import run_tests_in_session


def test_generator_fixture(suite, suite_test, get_fixture_location):
    fixture = get_fixture_location(suite_test).add_generator_fixture()
    suite_test.depend_on_fixture(fixture)
    suite.run()


def test_generator_fixture_with_name():

    @slash.generator_fixture(name='a_generator')
    def my_generator():
        yield 5

    @slash.generator_fixture
    def second_generator():  # pylint: disable=unused-variable
        yield 10

    def test_something(a_generator, second_generator):
        assert a_generator == 5
        assert second_generator == 10

    session = slash.Session()
    session.fixture_store.add_fixture(my_generator)
    session.fixture_store.add_fixture(second_generator)
    session.fixture_store.resolve()

    with session, session.get_started_context():
        run_tests_in_session(test_something, session=session)
    assert session.results.is_success(), 'run failed'



import sys
import gossip
import pytest
import inspect
import traceback
import slash
import logbook
from slash import exception_handling
from slash._compat import ExitStack, PYPY
from slash.exceptions import SkipTest, ExpectedExceptionNotCaught
from slash.utils import debug

from .utils import CustomException, TestCase


def test_handling_exceptions_swallow_skip_test(suite, suite_test):

    @suite_test.append_body
    def __code__():             # pylint: disable=unused-variable
        from slash.exception_handling import handling_exceptions
        with handling_exceptions(swallow=True):
            slash.skip_test()
        __ut__.events.add('NEVER') # pylint: disable=undefined-variable

    suite_test.expect_skip()

    summary = suite.run()
    assert not summary.events.has_event('NEVER')


def test_handling_exceptions_skip_test_would_not_return_skip_test():
    with pytest.raises(SkipTest) as caught:
        with slash.Session():
            with exception_handling.handling_exceptions() as handled:
                raise SkipTest('Unittest')

    assert isinstance(caught.value, SkipTest)
    assert handled.exception is caught.value
    assert exception_handling.is_exception_handled(caught.value)


def test_passthrough_types():

    value = CustomException()

    with slash.Session():
        with pytest.raises(CustomException) as caught:
            with exception_handling.handling_exceptions(passthrough_types=(CustomException,)) as handled:
                raise value
    assert value is caught.value
    assert handled.exception is None
    assert not exception_handling.is_exception_handled(value)

    with slash.Session():
        with pytest.raises(CustomException) as caught:
            with exception_handling.handling_exceptions(passthrough_types=(AttributeError,)) as handled:
                raise value
    assert value is caught.value
    assert handled.exception is value
    assert exception_handling.is_exception_handled(value)


def test_swallow_types():
    value = CustomException()

    with slash.Session():
        with exception_handling.handling_exceptions(swallow_types=(CustomException,)) as handled:
            raise value
    assert sys.exc_info() == exception_handling.NO_EXC_INFO
    assert handled.exception is value


class FakeTracebackTest(TestCase):
    def setUp(self):
        super(FakeTracebackTest, self).setUp()
        self.override_config("debug.enabled", True)
        self.forge.replace_with(debug, "launch_debugger", self.verify_fake_traceback_debugger)
        self.debugger_called = False
        self._tb_len = 0

    @pytest.mark.skipif(PYPY, reason='Cannot run on PyPy')
    def test_fake_traceback(self):
        with slash.Session(), pytest.raises(ZeroDivisionError):
            with exception_handling.handling_exceptions(fake_traceback=False):
                self._expected_line_number = inspect.currentframe().f_lineno + 1
                a = 1 / 0
                return a

        with slash.Session(), pytest.raises(ZeroDivisionError):
            with exception_handling.handling_exceptions():
                self._expected_line_number = inspect.currentframe().f_lineno + 1
                a = 1 / 0
                return a

    def verify_fake_traceback_debugger(self, exc_info):
        assert traceback.extract_tb(exc_info[2])[-1][1] == self._expected_line_number
        if not self._tb_len:
            # First attempt, no fake traceback
            self._tb_len = self._get_tb_len(exc_info[2])
            assert self._tb_len != 0
        else:
            # Second attempt, with fake traceback
            assert self._tb_len < self._get_tb_len(exc_info[2])

    def _get_tb_len(self, tb):
        tb_len = 0
        while tb:
            tb_len += 1
            tb = tb.tb_next
        return tb_len


def test_handling_exceptions():
    value = CustomException()

    with slash.Session(), pytest.raises(CustomException) as caught:
        with exception_handling.handling_exceptions() as handled1:
            with exception_handling.handling_exceptions() as handled2:
                with exception_handling.handling_exceptions() as handled3:
                    raise value

    assert caught.value is value
    assert handled1.exception is value
    assert handled2.exception is value
    assert handled3.exception is value


@pytest.mark.skipif(sys.version_info >= (3, 0), reason='Cannot run on 3.x')
def test_reraise_after_exc_info_reset():
    @gossip.register('slash.exception_caught_before_debugger')
    def exception_hook():       # pylint: disable=unused-variable
        sys.exc_clear()  # pylint: disable=no-member

    with slash.Session(), pytest.raises(CustomException):
        with exception_handling.handling_exceptions():
            raise CustomException()



class DebuggingTest(TestCase):

    def setUp(self):
        super(DebuggingTest, self).setUp()
        self.forge.replace_with(debug, "launch_debugger", self.dummy_debugger)
        self.debugger_called = False

    def dummy_debugger(self, *args, **kwargs):  # pylint: disable=unused-argument
        self.debugger_called = True

    def test_debugging_not_configured(self):
        self._raise_exception_in_context(ZeroDivisionError)
        self.assertFalse(self.debugger_called)

    def test_debugging_configured_no_skips(self):
        self.override_config("debug.debug_skips", False)
        self.override_config("debug.enabled", True)
        self._raise_exception_in_context(SkipTest)
        self.assertFalse(self.debugger_called)

    def test_debugging_skips(self):
        self.override_config("debug.debug_skips", True)
        self.override_config("debug.enabled", True)
        self._raise_exception_in_context(SkipTest)
        self.assertTrue(self.debugger_called)

    def _raise_exception_in_context(self, exception_type):
        with slash.Session(), self.assertRaises(exception_type):
            with exception_handling.handling_exceptions():
                raise exception_type()


def test_swallow_exceptions():
    with exception_handling.get_exception_swallowing_context():
        raise CustomException("!!!")


def test_no_swallow():
    raised = CustomException()
    with pytest.raises(CustomException) as caught:
        with exception_handling.get_exception_swallowing_context():
            raise exception_handling.noswallow(raised)
    assert raised is caught.value


def test_disable_exception_swallowing_function():
    raised = CustomException()
    with pytest.raises(CustomException) as caught:
        with exception_handling.get_exception_swallowing_context():
            exception_handling.disable_exception_swallowing(raised)
            raise raised
    assert caught.value is raised


def test_disable_exception_swallowing_decorator():
    raised = CustomException()
    @exception_handling.disable_exception_swallowing
    def func():
        raise raised
    with pytest.raises(CustomException) as caught:
        with exception_handling.get_exception_swallowing_context():
            func()
    assert caught.value is raised


@pytest.mark.parametrize('message', [None, 'My custom message'])
@pytest.mark.parametrize('exc_types', [CustomException, (CustomException, ZeroDivisionError)])
def test_assert_raises(exc_types, message):
    raised = CustomException()
    with slash.Session():
        with slash.assert_raises(exc_types, msg=message) as caught:
            raise raised
    assert sys.exc_info() == exception_handling.NO_EXC_INFO
    assert caught.exception is raised


@pytest.mark.parametrize('message', [None, 'My custom message'])
def test_assert_raises_that_not_raises(message):
    expected_substring = message or 'not raised'
    try:
        with slash.assert_raises(CustomException, msg=message):
            pass
    except ExpectedExceptionNotCaught as e:
        assert expected_substring in str(e)
        assert e.expected_types == (CustomException,)
    else:
        raise Exception('ExpectedExceptionNotCaught exception was not raised :(')


@pytest.mark.parametrize('with_session', [True, False])
def test_handling_exceptions_inside_assert_raises_with_session(with_session):
    value = CustomException()

    with ExitStack() as ctx:

        if with_session:
            session = ctx.enter_context(slash.Session())
            ctx.enter_context(session.get_started_context())
        else:
            session = None

        with slash.assert_raises(CustomException):
            with exception_handling.handling_exceptions():
                raise value

    assert not exception_handling.is_exception_handled(value)
    if with_session:
        assert session.results.get_num_errors() == 0


@pytest.mark.parametrize('message', [None, 'My custom message'])
@pytest.mark.parametrize('exc_types', [CustomException, (CustomException, ZeroDivisionError)])
def test_allowing_exceptions(exc_types, message):
    raised = CustomException()
    with slash.Session():
        with slash.allowing_exceptions(exc_types, msg=message) as caught:
            with logbook.TestHandler() as handler:
                raise raised
    assert sys.exc_info() == exception_handling.NO_EXC_INFO
    assert caught.exception is raised
    assert len(handler.records) == 0  # pylint: disable=len-as-condition


def test_allowing_exceptions_that_not_raises():
    with logbook.TestHandler() as handler:
        with slash.allowing_exceptions((Exception, CustomException)):
            pass
    assert len(handler.records) == 1
    assert handler.records[0].message.startswith("Exception/CustomException not raised")


@pytest.mark.parametrize('with_session', [True, False])
def test_handling_exceptions_inside_allowing_exceptions_with_session(with_session):
    value = CustomException()

    with ExitStack() as ctx:

        if with_session:
            session = ctx.enter_context(slash.Session())
            ctx.enter_context(session.get_started_context())
        else:
            session = None

        with logbook.TestHandler() as handler:
            with slash.allowing_exceptions(CustomException):
                with exception_handling.handling_exceptions():
                    raise value

    assert not exception_handling.is_exception_handled(value)
    if with_session:
        assert session.results.get_num_errors() == 0
    assert len(handler.records) == 0  # pylint: disable=len-as-condition



from __future__ import absolute_import

from slash import config as slash_config

from ...utils.conf_utils import Cmdline, Doc
from ...utils import parallel_utils
from ..interface import PluginInterface

_DATA_FILENAME = '.coverage'

class Plugin(PluginInterface):

    """Enables saving coverage information for test runs
    For more information see https://slash.readthedocs.org/en/master/builtin_plugins.html#coverage
    """

    def get_name(self):
        return "coverage"

    def get_default_config(self):
        return {
            'config_filename': False // Cmdline(arg='--cov-config') // Doc('Coverage configuration file'),
            'report_type': 'html' // Cmdline(arg='--cov-report') // Doc('Coverage report format'),
            'report': True,
            'append': False // Cmdline(on='--cov-append') // Doc('Append coverage data to existing file'),
            'sources': [] // Cmdline(append='--cov') // Doc('Modules or packages for which to track coverage'),
        }

    def activate(self):
        try:
            import coverage
        except ImportError: # pragma: no cover
            raise RuntimeError('The coverage plugin requires the coverage package to be installed. Please run `pip install coverage` to install it')

        sources = slash_config.root.plugin_config.coverage.sources or None
        data_file_name = _DATA_FILENAME
        if parallel_utils.is_child_session():
            data_file_name += "." + slash_config.root.parallel.worker_id
        self._cov = coverage.Coverage(
            data_file=data_file_name,
            config_file=slash_config.root.plugin_config.coverage.config_filename,
            source=sources,
        )
        if slash_config.root.plugin_config.coverage.append:
            self._cov.load()
        self._reporters = []
        for report_type_name in slash_config.root.plugin_config.coverage.report_type.split(','):
            if report_type_name == 'html':
                self._reporters.append(self._cov.html_report)
            elif report_type_name == 'xml':
                self._reporters.append(self._cov.xml_report)
            else:
                raise RuntimeError('Unknown report type: {!r}'.format(report_type_name))
        self._cov.start()


    def session_end(self):
        from coverage import CoverageException
        self._cov.stop()
        self._cov.save()
        if slash_config.root.plugin_config.coverage.report:
            for reporter in self._reporters:
                try:
                    if parallel_utils.is_child_session():
                        continue
                    elif parallel_utils.is_parent_session():
                        self._cov.combine()
                    reporter()
                except CoverageException as e:
                    if 'no data' not in str(e).lower():
                        raise



from contextlib import contextmanager

import logbook

from . import hooks
from . import log
from ._compat import ExitStack
from .conf import config
from .ctx import context
from .exception_handling import handling_exceptions
from .exceptions import NoActiveSession
from .core.function_test import FunctionTest
from .core.metadata import ensure_test_metadata
from .core.exclusions import is_excluded
from .core import requirements
from .utils.iteration import PeekableIterator


_logger = logbook.Logger(__name__)
log.set_log_color(_logger.name, logbook.NOTICE, 'blue')


def run_tests(iterable, stop_on_error=None):
    """
    Runs tests from an iterable using the current session
    """
    # pylint: disable=maybe-no-member
    def should_stop_on_error():
        if stop_on_error is None:
            return config.root.run.stop_on_error
        return stop_on_error

    if context.session is None or not context.session.started:
        raise NoActiveSession("A session is not currently started")

    test_iterator = PeekableIterator(iterable)
    last_filename = None
    complete = False
    try:
        for test in test_iterator:
            if config.root.run.dump_variation:
                _dump_variation(test)
            _set_test_metadata(test)
            test_filename = test.__slash__.file_path
            if last_filename != test_filename:
                context.session.reporter.report_file_start(test_filename)
                last_filename = test_filename
            context.session.reporter.report_test_start(test)
            _logger.notice(
                "#{}: {}",
                test.__slash__.test_index1,
                test.__slash__.get_address(
                    raw_params=config.root.log.show_raw_param_values
                ),
                extra={'highlight': True, 'filter_bypass': True})

            _run_single_test(test, test_iterator)


            result = context.session.results[test]
            context.session.reporter.report_test_end(test, result)
            if not test_iterator.has_next() or ensure_test_metadata(test_iterator.peek()).file_path != last_filename:
                context.session.reporter.report_file_end(last_filename)
            if result.has_fatal_exception():
                _logger.debug("Stopping on fatal exception")
                break
            if should_stop_on_error() and not context.session.results.is_success(allow_skips=True):
                _logger.debug("Stopping (run.stop_on_error==True)")
                break
        else:
            complete = True
    finally:
        if config.root.parallel.worker_id is None:
            context.session.initiate_cleanup()

    if config.root.parallel.worker_id is None:
        _mark_unrun_tests(test_iterator)
        if complete:
            context.session.mark_complete()
        elif last_filename is not None:
            context.session.reporter.report_file_end(last_filename)
        _logger.trace('Session finished. is_success={0} has_skips={1}',
                  context.session.results.is_success(allow_skips=True), bool(context.session.results.get_num_skipped()))


def _dump_variation(test):
    _logger.trace('Variation information:\n{}',
                  '\n'.join('\t{}: {!r}'.format(k, v) for k, v in sorted(test.get_variation().id.items())))


def _run_single_test(test, test_iterator):
    next_test = test_iterator.peek_or_none()
    with ExitStack() as exit_stack:

        # sets the current result, test id etc.
        result, prev_result = exit_stack.enter_context(_get_test_context(test))

        with handling_exceptions():


            should_run = _process_requirements_and_exclusions(test)
            if not should_run:
                return

            result.mark_started()
            with TestStartEndController(result, prev_result) as controller:
                try:
                    try:
                        with handling_exceptions(swallow=True):
                            context.session.scope_manager.begin_test(test)
                            try:
                                controller.start()
                                with handling_exceptions(swallow=True):
                                    test.run()
                            finally:
                                context.session.scope_manager.end_test(test)
                    except context.session.get_skip_exception_types():
                        pass

                    _fire_test_summary_hooks(test, result)
                    if next_test is None and config.root.parallel.worker_id is None:
                        controller.end()

                        with handling_exceptions(swallow=True):
                            context.session.initiate_cleanup()

                except context.session.get_skip_exception_types():
                    pass

def _process_requirements_and_exclusions(test):
    """Returns whether or not a test should run based on requirements and exclusions, also triggers skips and relevant hooks
    """
    unmet_reqs = test.get_unmet_requirements()
    if not unmet_reqs:
        return _process_exclusions(test)


    messages = set()
    for req, message in unmet_reqs:
        if isinstance(req, requirements.Skip):
            context.result.add_skip(req.reason)
            msg = 'Skipped' if not req.reason else req.reason
        else:
            msg = 'Unmet requirement: {}'.format(message or req)
            context.result.add_skip(msg)
        messages.add(msg)

    hooks.test_avoided(reason=', '.join(messages)) # pylint: disable=no-member
    return False

def _process_exclusions(test):
    if is_excluded(test):
        context.result.add_skip('Excluded')
        hooks.test_avoided(reason='Excluded') # pylint: disable=no-member
        return False
    return True


class TestStartEndController(object):

    def __init__(self, result, prev_result):
        self._result = result
        self._prev_result = prev_result
        self._started = False

    def __enter__(self):
        return self

    def start(self):
        if not self._started:
            self._started = True
            self._result.mark_started()
            hooks.test_start() # pylint: disable=no-member

    def end(self):
        if self._started:
            self._started = False
            try:
                with context.session.cleanups.forbid_implicit_scoping_context():
                    hooks.test_end() # pylint: disable=no-member
                    self._result.mark_finished()
            finally:
                context.result = self._prev_result


    def __exit__(self, *args):
        self.end()


def _fire_test_summary_hooks(test, result): # pylint: disable=unused-argument
    with handling_exceptions():
        if result.is_just_failure():
            hooks.test_failure()  # pylint: disable=no-member
        elif result.is_skip():
            hooks.test_skip(reason=result.get_skips()[0]) # pylint: disable=no-member
        elif result.is_success():
            hooks.test_success()  # pylint: disable=no-member
        else:
            _logger.debug('Firing test_error hook for {0} (result: {1})', test, result)
            hooks.test_error()  # pylint: disable=no-member


def _set_test_metadata(test):
    ensure_test_metadata(test)
    assert test.__slash__.test_index0 is None
    test.__slash__.test_index0 = next(context.session.test_index_counter) if config.root.parallel.worker_id is None \
                                 else context.session.current_parallel_test_index


def _mark_unrun_tests(test_iterator):
    remaining = list(test_iterator)
    for test in remaining:
        with _get_test_context(test, logging=False):
            pass

@contextmanager
def _get_test_context(test, logging=True):
    ensure_test_metadata(test)

    with _set_current_test_context(test):
        result = context.session.results.create_result(test)
        prev_result = context.result
        context.result = result
        try:
            with (context.session.logging.get_test_logging_context(result) if logging else ExitStack()):
                _logger.debug("Started test #{0.__slash__.test_index1}: {0}", test)
                yield result, prev_result
        finally:
            context.result = prev_result

@contextmanager
def _set_current_test_context(test):
    prev_test = context.test
    prev_test_id = context.test_id
    context.test = test
    context.test_id = test.__slash__.id
    if isinstance(test, FunctionTest):
        context.test_classname = None
        context.test_methodname = test.__slash__.factory_name
    else:
        context.test_classname = test.__slash__.factory_name
        # this includes a dot (.), so it has to be truncated
        context.test_methodname = test.__slash__.address_in_factory[1:]
    try:
        yield
    finally:
        context.test = prev_test
        context.test_id = prev_test_id



from collections import OrderedDict
import functools
import warnings
import inspect
import sys
import ast
import pickle

from sentinels import NOTHING

from .._compat import reraise, PY2, PYPY

def check_duplicate_functions(path):
    code = None
    with open(path) as f:
        code = f.read()
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        root = ast.parse(code, filename=path)
    func_names = set()
    duplicates = set()
    for node in root.body:
        if isinstance(node, ast.FunctionDef):
            if node.name in func_names:
                duplicates.add((path, node.name, node.lineno))
            else:
                func_names.add(node.name)
    return duplicates


def wraps(func, preserve=()):
    def decorator(new_func):
        returned = functools.wraps(func)(new_func)
        returned.__wraps__ = func
        return returned
    for p in preserve:
        orig = getattr(func, p, NOTHING)
        if orig is not NOTHING:
            setattr(decorator, p, orig)
    return decorator

def unpickle(thing):
    if not thing:
        return thing
    if PY2:
        return pickle.loads(thing)
    else:
        return pickle.loads(thing.data)

def get_underlying_func(func):
    while True:
        underlying = getattr(func, "__wraps__", None)
        if underlying is None:
            return func
        func = underlying


def get_argument_names(func):
    return [arg.name for arg in get_arguments(func)]


def get_arguments_dict(func):
    returned = OrderedDict()
    for arg in get_arguments(func):
        returned[arg.name] = arg
    return returned


def get_arguments(func):
    if PY2 or PYPY:
        func = get_underlying_func(func)
        spec = inspect.getargspec(func)  # pylint: disable=deprecated-method
        returned = [FunctionArgument(name=name) for name in spec.args]
    else:
        if getattr(func, '__self__', None) is not None:
            func = func.__func__ # signature() doesn't work on bound methods
        returned = [FunctionArgument.from_parameter(p) for name, p in inspect.signature(func).parameters.items()] # pylint: disable=no-member

    if returned and returned[0].name == 'self':
        returned = returned[1:]
    return returned


class FunctionArgument(object):

    def __init__(self, name, annotation=NOTHING):
        super(FunctionArgument, self).__init__()
        self.name = name
        self.annotation = annotation

    @classmethod
    def from_parameter(cls, parameter):
        annotation = parameter.annotation
        if annotation is parameter.empty:
            annotation = NOTHING
        return cls(name=parameter.name, annotation=parameter.annotation)


def call_all_raise_first(_funcs, *args, **kwargs):
    exc_info = None
    for func in _funcs:
        try:
            func(*args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            exc_info = sys.exc_info()
    if exc_info is not None:
        reraise(*exc_info)



import colorama
import pytest
from slash._compat import StringIO
from slash.utils.cli_utils import Printer, make_styler, UNDERLINED

_style_1 = colorama.Fore.MAGENTA + colorama.Style.BRIGHT + UNDERLINED  # pylint: disable=no-member
_style_2 = colorama.Fore.GREEN + colorama.Style.BRIGHT  # pylint: disable=no-member

def _toggle(toggle_name):
    return pytest.mark.parametrize(toggle_name, [True, False])

def _colorized(string, style=None):
    if not style:
        return string
    return make_styler(style)(string)

@_toggle('force_color')
@_toggle('enable_color')
def test_printer_with_output_disabled(force_color, enable_color):
    report_stream = StringIO()
    printer = Printer(report_stream, enable_output=False, force_color=force_color, enable_color=enable_color)
    printer(_colorized('A', _style_1))
    printer(_colorized('B', _style_2))
    printer('C')
    assert not report_stream.getvalue()


def test_printer_with_forced_colored():
    report_stream = StringIO()
    printer = Printer(report_stream, force_color=True)
    expected_lines = []
    for string, style in [('A', _style_1), ('B', _style_2), ('C', None)]:
        colored_string = _colorized(string, style)
        printer(colored_string)
        expected_lines.append(colored_string.colorize() if style else str(colored_string))
    assert report_stream.getvalue().splitlines() == expected_lines


def test_printer_with_disalbed_colored():
    report_stream = StringIO()
    printer = Printer(report_stream, enable_color=True)
    printer(_colorized('A', _style_1))
    printer(_colorized('B', _style_2))
    printer('C')
    assert report_stream.getvalue().splitlines() == ['A', 'B', 'C']



import logbook
import pytest

from slash._compat import StringIO
from slash.reporting.console_reporter import ConsoleReporter


@pytest.mark.parametrize('console_level', [True, False])  # pylint: disable=redefined-outer-name
def test_console_reporter(suite, level, config_override, console_level):  # pylint: disable=redefined-outer-name
    config_override('log.color_console', console_level)
    config_override('log.console_level', level)
    summary = suite.run()
    suite.add_test().when_run.raise_exception()
    output = summary.get_console_output()
    assert output
    assert bool(console_level) == ('\x1b' in output)


def test_silence_manual_errors(suite, suite_test, config_override):
    suite_test.append_line('slash.add_error("msg")')
    suite_test.expect_error()
    config_override('log.show_manual_errors_tb', False)
    output = suite.run().get_console_output()
    assert 'slash.add_error' not in output


def test_num_collected_printed_once(suite, suite_test):  # pylint: disable=unused-argument
    assert len(suite) > 1
    output = suite.run().get_console_output()
    assert output.count('collected') == 1


@pytest.mark.parametrize('long_headline', [True, False])
@pytest.mark.parametrize('multiline', [True, False])
def test_fancy_message(long_headline, multiline):
    output = StringIO()
    reporter = ConsoleReporter(logbook.TRACE, output)
    headline = 'some headline here'
    if long_headline:
        headline *= 80
    message = 'some message here'

    repetitions = 5
    if multiline:
        message = "\n".join(message for i in range(repetitions))
        message += '\n\n'

    reporter.report_fancy_message(headline, message)
    if long_headline:
        assert headline[:80] in output.getvalue()
    else:
        assert headline in output.getvalue()

    if multiline:
        assert output.getvalue().count(message.splitlines()[0]) == repetitions
    else:
        assert output.getvalue().count(message) == 1


@pytest.fixture(params=list(range(logbook.DEBUG, logbook.CRITICAL + 1)))
def level(request):
    return request.param


@pytest.fixture
def stream():
    return StringIO()



import pytest
import slash
import time
import logbook
from slash.utils.interactive import notify_if_slow_context
from slash.reporting.console_reporter import ConsoleReporter
from slash._compat import StringIO


@pytest.mark.parametrize('show_duration', [True, False])
def test_notify_if_slow_context(show_duration):
    output_stream = StringIO()
    reporter = ConsoleReporter(logbook.TRACE, output_stream)
    with slash.Session(console_stream=output_stream, reporter=reporter):
        with notify_if_slow_context('message', slow_seconds=0.1, end_message='End', show_duration=show_duration):
            time.sleep(1)
    output = output_stream.getvalue()
    assert 'message' in output
    assert 'End' in output
    assert ('took 0:00:01' in output) == show_duration



import operator
import sys

import logbook
from vintage import deprecated

from . import exception_handling
from ._compat import PY2
from .exceptions import TestFailed, ExpectedExceptionNotCaught
from .utils import operator_information

sys.modules["slash.should"] = sys.modules[__name__]
_logger = logbook.Logger(__name__)

def _deprecated(func, message=None):
    return deprecated(since='0.19.0', what='slash.should.{.__name__}'.format(func),
                      message=message or 'Use plain assertions instead')(func)


def _binary_assertion(name, operator_func):
    op = operator_information.get_operator_by_func(operator_func)

    def _assertion(a, b, msg=None):
        if not op(a, b):
            msg = _get_message(msg, operator_information.get_operator_by_func(
                op.inverse_func).to_expression(a, b))
            raise TestFailed(msg)
    _assertion.__name__ = name
    _assertion.__doc__ = "Asserts **{}**".format(
        op.to_expression("ARG1", "ARG2"))
    _assertion = _deprecated(_assertion)
    return _assertion


def _unary_assertion(name, operator_func):
    op = operator_information.get_operator_by_func(operator_func)

    def _assertion(a, msg=None):
        if not op(a):
            msg = _get_message(msg, operator_information.get_operator_by_func(
                op.inverse_func).to_expression(a))
            raise TestFailed(msg)
    _assertion.__name__ = name
    _assertion.__doc__ = "Asserts **{}**".format(op.to_expression("ARG"))
    _assertion = _deprecated(_assertion)
    return _assertion


def _get_message(msg, description):
    if msg is None:
        return description
    return "{} ({})".format(msg, description)

equal = _binary_assertion("equal", operator.eq)
assert_equal = assert_equals = equal = equal

not_equal = _binary_assertion("not_equal", operator.ne)
assert_not_equal = assert_not_equals = not_equals = not_equal

be_a = _binary_assertion("be_a", operator_information.safe_isinstance)
assert_isinstance = be_a

not_be_a = _binary_assertion(
    "not_be_a", operator_information.safe_not_isinstance)
assert_not_isinstance = not_be_a

be_none = _unary_assertion("be_none", operator_information.is_none)
assert_is_none = be_none

not_be_none = _unary_assertion("not_be_none", operator_information.is_not_none)
assert_is_not_none = not_be_none

be = _binary_assertion("be", operator.is_)
assert_is = be

not_be = _binary_assertion("not_be", operator.is_not)
assert_is_not = not_be

be_true = _unary_assertion("be_true", operator.truth)
assert_true = be_true

be_false = _unary_assertion("be_false", operator.not_)
assert_false = be_false

be_empty = _unary_assertion("be_empty", operator_information.is_empty)
assert_empty = assert_is_empty = be_empty

not_be_empty = _unary_assertion(
    "not_be_empty", operator_information.is_not_empty)
assert_not_empty = assert_is_not_empty = not_be_empty

contain = _binary_assertion("contain", operator.contains)
assert_contains = contains = contain

not_contain = _binary_assertion(
    "not_contain", operator_information.not_contains)
assert_not_contains = assert_not_contain = not_contains = not_contain


def be_in(a, b, msg=None):
    """
    Asserts **ARG1 in ARG2**
    """
    return contain(b, a, msg)
assert_in = be_in


def not_be_in(a, b, msg=None):
    """
    Asserts **ARG1 not in ARG2**
    """
    return not_contain(b, a, msg)
assert_not_in = not_be_in


class _CaughtContext(object):

    def __init__(self, message, exc_types, ensure_caught):
        if not isinstance(exc_types, tuple):
            exc_types = (exc_types, )
        self._expected_classes = exc_types
        self._caught = _CaughtException()
        self._ignore_ctx = None
        self._msg = message
        self._ensure_caught = ensure_caught

    def __enter__(self):
        self._ignore_ctx = exception_handling.thread_ignore_exception_context(self._expected_classes)
        self._ignore_ctx.__enter__()  # pylint: disable=no-member
        return self._caught

    def __exit__(self, *exc_info):
        if self._ignore_ctx:
            self._ignore_ctx.__exit__(*exc_info)  # pylint: disable=no-member
        if exc_info and exc_info != exception_handling.NO_EXC_INFO:
            e = exc_info[1]
            if isinstance(e, self._expected_classes):
                self._caught.exception = e
                if PY2:
                    sys.exc_clear()  # pylint: disable=no-member
                return True
            return None
        msg = self._msg
        if self._msg is None:
            expected_classes = self._expected_classes
            if not isinstance(expected_classes, tuple):
                expected_classes = (expected_classes, )
            msg = "{} not raised".format("/".join(e.__name__ for e in expected_classes))
        if self._ensure_caught:
            raise ExpectedExceptionNotCaught(msg, self._expected_classes)
        _logger.debug(msg)
        return True


def assert_raises(exception_class, msg=None):
    """
    Ensures a subclass of **ARG1** leaves the wrapped context:

    >>> with assert_raises(AttributeError):
    ...     raise AttributeError()
    """
    return _CaughtContext(msg, exception_class, ensure_caught=True)


def allowing_exceptions(exception_class, msg=None):
    """
    Allow subclass of **ARG1** to be raised during context:

    >>> with allowing_exceptions(AttributeError):
    ...     raise AttributeError()
    >>> with allowing_exceptions(AttributeError):
    ...     pass
    """
    return _CaughtContext(msg, exception_class, ensure_caught=False)


@deprecated(since='0.19.0', what='slash.should.raise_exception', message='Use slash.assert_raises instead')
def raise_exception(exception_class, msg=None):
    return assert_raises(exception_class, msg=msg)

raise_exception.__doc__ = assert_raises.__doc__.replace("assert_raises", "raise_exception")


def assert_almost_equal(a, b, delta=0.00000001):
    """Asserts that abs(a - b) <= delta
    """
    assert abs(a - b) <= delta


class _CaughtException(object):
    exception = None



import collections
import gossip
import itertools
import os
import re
import sys
from .interface import PluginInterface
from .registration_info import RegistrationInfo
from .._compat import itervalues, reraise
from ..utils.marks import try_get_mark
from ..utils import parallel_utils
from ..conf import config
from contextlib import contextmanager
from emport import import_file
from sentinels import NOTHING
from vintage import warn_deprecation
import logbook

_logger = logbook.Logger(__name__)

_SKIPPED_PLUGIN_METHOD_NAMES = set(dir(PluginInterface))
PluginInfo = collections.namedtuple("PluginInfo", ("plugin_instance", "is_internal"))
_DEPRECATED_CHARACTERS = '-_'

class IncompatiblePlugin(ValueError):
    pass

class UnknownPlugin(ValueError):
    pass

class IllegalPluginName(ValueError):
    pass

class PluginManager(object):
    def __init__(self):
        super(PluginManager, self).__init__()
        self._installed = {}
        self._cmd_line_to_name = {}
        self._config_to_name = {}
        self._pending_activation = set()
        self._pending_deactivation = set()
        self._active = set()
        self.install_builtin_plugins()

    @contextmanager
    def restoring_state_context(self):
        previous_installed = self._installed.copy()
        previous_active = {name: self._installed[name] for name in self._active}
        try:
            yield
        finally:
            for name in set(previous_installed) - set(self._installed):
                self.install(previous_installed[name].plugin_instance)
            for name in set(previous_active) - self._active:
                self.activate(name)
            for name in self._active - set(previous_active):
                self.deactivate(name)
            for name in set(self._installed) - set(previous_installed):
                self.uninstall(name)


    def discover(self):
        """
        Iterates over all search paths and loads plugins
        """
        for search_path in config.root.plugins.search_paths:
            for path, _, filenames in os.walk(search_path):
                for filename in filenames:
                    if not filename.endswith(".py"):
                        continue
                    module = import_file(os.path.join(path, filename))
                    install_func = getattr(module, "install_plugins", None)
                    if install_func is None:
                        continue
                    install_func()

    def get_installed_plugins(self, include_internals=True):
        """
        Returns a dict mapping plugin names to currently installed plugins
        """
        return {plugin_name: plugin_info.plugin_instance
                for plugin_name, plugin_info in self._installed.items()
                if include_internals or (not plugin_info.is_internal)}

    def get_active_plugins(self):
        """
        Returns a dict mapping plugin names to currently active plugins
        """
        return dict(self._iterate_active_plugins())

    def _iterate_active_plugins(self):
        for active_name in self._active:
            yield (active_name, self._get_installed_plugin_instance_by_name(active_name))

    def get_future_active_plugins(self):
        """
        Returns a dictionary of plugins intended to be active once the 'pending activation' mechanism
        is finished
        """
        returned = self.get_active_plugins()
        for name in self._pending_activation:
            returned[name] = self.get_plugin(name)
        for name in self._pending_deactivation:
            returned.pop(name, None)
        return returned

    def get_plugin(self, plugin_name):
        """
        Retrieves a registered plugin by name, or raises a LookupError
        """
        return self._installed[plugin_name].plugin_instance

    def is_internal_plugin(self, plugin):
        """
        Returns rather installed plugin is internal plugin
        """
        plugin_name = plugin if isinstance(plugin, str) else plugin.get_name()
        return self._installed[plugin_name].is_internal

    def _is_parallel_supported(self, plugin):
        if not parallel_utils.is_parallel_session():
            return False
        plugin_parallel_mode = try_get_mark(plugin, 'parallel_mode', parallel_utils.ParallelPluginModes.ENABLED)
        if plugin_parallel_mode == parallel_utils.ParallelPluginModes.ENABLED:
            return False
        if (plugin_parallel_mode == parallel_utils.ParallelPluginModes.DISABLED) \
            or (plugin_parallel_mode == parallel_utils.ParallelPluginModes.PARENT_ONLY and parallel_utils.is_child_session()) \
            or (plugin_parallel_mode == parallel_utils.ParallelPluginModes.CHILD_ONLY and parallel_utils.is_parent_session()):
            return True
        return False

    def configure_for_parallel_mode(self):
        for plugin in self.get_installed_plugins().values():
            if self._is_parallel_supported(plugin):
                self.deactivate_later(plugin)

    def install(self, plugin, activate=False, activate_later=False, is_internal=False):
        """
        Installs a plugin object to the plugin mechanism. ``plugin`` must be an object deriving from
        :class:`slash.plugins.PluginInterface`.
        """
        if not isinstance(plugin, PluginInterface):
            raise IncompatiblePlugin("Invalid plugin type: {!r}".format(type(plugin)))
        plugin_name = plugin.get_name()
        if re.search(r'[^A-Za-z0-9_ -]', plugin_name):
            raise IllegalPluginName("Illegal plugin name: {}".format(plugin_name))

        if any(char in plugin_name for char in _DEPRECATED_CHARACTERS):
            warn_deprecation("In the future, dashes and underscore will not be allowed in plugin names - "
                             "spaces should be used instead (plugin name: {!r})".format(plugin_name))
        self._configure(plugin)
        self._installed[plugin_name] = PluginInfo(plugin, is_internal)
        self._cmd_line_to_name[self.normalize_command_line_name(plugin_name)] = plugin_name
        self._config_to_name[self.normalize_config_name(plugin_name)] = plugin_name
        if not hasattr(plugin, '__toggles__'):
            plugin.__toggles__ = {
                'session': gossip.Toggle(),
            }
        if activate:
            try:
                self.activate(plugin_name)
            except IncompatiblePlugin:
                exc_info = sys.exc_info()
                self.uninstall(plugin)
                reraise(*exc_info)
        if activate_later:
            self.activate_later(plugin_name)

    def install_builtin_plugins(self):
        for builtin_plugin_module in self._iter_builtin_plugin_modules():
            module = __import__(
                "slash.plugins.builtin.{}".format(builtin_plugin_module),
                fromlist=[""]
            )
            self.install(module.Plugin())

    def _iter_builtin_plugin_modules(self):
        builtin_dir = os.path.join(os.path.dirname(__file__), "builtin")
        for filename in os.listdir(builtin_dir):
            if filename.startswith("_") or filename.startswith(".") or not filename.endswith(".py"):
                continue
            yield filename[:-3]

    def uninstall(self, plugin):
        """
        Uninstalls a plugin
        """
        plugin = self._get_installed_plugin(plugin)
        try:
            self.deactivate(plugin)
        except IncompatiblePlugin:
            pass
        self._unconfigure(plugin)
        plugin_name = plugin.get_name()
        self._installed.pop(plugin_name)
        cmd_name = self.normalize_command_line_name(plugin_name)
        self._cmd_line_to_name.pop(cmd_name, None)
        config_name = self.normalize_config_name(plugin_name)
        self._config_to_name.pop(config_name, None)

    def uninstall_all(self):
        """
        Uninstalls all installed plugins
        """
        for plugin_info in list(itervalues(self._installed)):
            self.uninstall(plugin_info.plugin_instance)
        assert not self._installed

    def activate(self, plugin):
        """
        Activates a plugin, registering its hook callbacks to their respective hooks.

        :param plugin: either a plugin object or a plugin name
        """
        plugin = self._get_installed_plugin(plugin)
        plugin_name = plugin.get_name()
        if self._is_parallel_supported(plugin):
            _logger.warn("Activating plugin {} though it's configuration for parallel mode doesn't fit to current session".format(plugin.get_name()))
        plugin.activate()
        for hook, callback, kwargs in self._get_plugin_registrations(plugin):
            hook.register(callback, **kwargs)
        self._active.add(plugin_name)

    def activate_later(self, plugin):
        """
        Adds a plugin to the set of plugins pending activation. It can be remvoed from the queue with :meth:`.deactivate_later`

        .. seealso:: :meth:`.activate_pending_plugins`
        """
        self._pending_activation.add(self._get_installed_plugin(plugin).get_name())

    def deactivate_later(self, plugin):
        """
        Removes a plugin from the set of plugins pending activation.

        .. seealso:: :meth:`.activate_pending_plugins`
        """

        self._pending_deactivation.add(self._get_installed_plugin(plugin).get_name())

    def activate_pending_plugins(self):
        """
        Activates all plugins queued with :meth:`.activate_later`
        """
        while self._pending_activation:
            plugin_name = self._pending_activation.pop()
            if plugin_name not in self._pending_deactivation:
                self.activate(plugin_name)

        while self._pending_deactivation:
            plugin_name = self._pending_deactivation.pop()
            if plugin_name in self._active:
                self.deactivate(plugin_name)

    def normalize_command_line_name(self, plugin_name):
        return plugin_name.replace(' ', '-')

    def normalize_config_name(self, plugin_name):
        return plugin_name.replace(' ', '_')

    def deactivate(self, plugin):
        """
        Deactivates a plugin, unregistering all of its hook callbacks

        :param plugin: either a plugin object or a plugin name
        """
        plugin = self._get_installed_plugin(plugin)
        plugin_name = plugin.get_name()
        token = self._get_token(plugin_name)

        if plugin_name in self._active:
            gossip.get_global_group().unregister_token(token)
            self._active.discard(plugin_name)
            plugin.deactivate()

    def _configure(self, plugin):
        cfg = plugin.get_config()
        if cfg is not None:
            warn_deprecation('PluginInterface.get_config() is deprecated. '
                             'Please use PluginInterface.get_default_config() instead')
        else:
            cfg = plugin.get_default_config()
        if cfg is not None:
            plugin_name = plugin.get_name()
            config_name = self.normalize_config_name(plugin_name)
            config['plugin_config'].extend({config_name: cfg})

    def _unconfigure(self, plugin):
        plugin_config = config['plugin_config']
        config_name = self.normalize_config_name(plugin.get_name())
        if config_name in plugin_config:
            plugin_config.pop(config_name)

    def _get_token(self, plugin_name):
        return "slash.plugins.{}".format(plugin_name)

    def _get_installed_plugin_instance_by_name(self, plugin_name):
        plugin_info = self._installed.get(plugin_name)
        if plugin_info is None:
            return None
        return plugin_info.plugin_instance

    def _get_installed_plugin_instance_by_type(self, plugin_class):
        for plugin in self._installed.values():
            if type(plugin.plugin_instance) is plugin_class: # pylint: disable=unidiomatic-typecheck
                return plugin.plugin_instance
        return None

    def _get_installed_plugin(self, plugin):
        if isinstance(plugin, str):
            plugin_name = plugin
            if plugin_name in self._cmd_line_to_name:
                plugin_name = self._cmd_line_to_name[plugin_name]
            plugin_instance = self._get_installed_plugin_instance_by_name(plugin_name)
        elif isinstance(plugin, type):
            plugin_instance = self._get_installed_plugin_instance_by_type(plugin)
            plugin_name = plugin_instance.get_name() if plugin_instance is not None else repr(plugin)
        else:
            plugin_instance = plugin
            plugin_name = plugin.get_name()
        if plugin_instance is None or self._get_installed_plugin_instance_by_name(plugin_name) is not plugin_instance:
            raise UnknownPlugin("Unknown plugin: {}".format(plugin_name))
        return plugin_instance

    def _get_plugin_registrations(self, plugin):
        plugin_name = plugin.get_name()
        returned = []
        unknown = []
        global_needs = try_get_mark(plugin, 'plugin_needs', [])
        global_provides = try_get_mark(plugin, 'plugin_provides', [])

        has_session_end = has_session_start = False

        register_no_op_hooks = set()
        if global_provides:
            register_no_op_hooks.update(hook.full_name for hook in gossip.get_group('slash').get_hooks())

        for method_name in dir(type(plugin)):
            if method_name in _SKIPPED_PLUGIN_METHOD_NAMES:
                continue

            method = getattr(plugin, method_name)

            if not hasattr(method, '__call__'):
                continue

            registration_list = try_get_mark(method, 'register_on', NOTHING)

            if registration_list is not NOTHING:
                registration_list = registration_list[:]
            else:
                if method_name.startswith('_'):
                    continue
                registration_list = [RegistrationInfo("slash.{}".format(method_name), expect_exists=True)]

            for registration_info in registration_list:
                if registration_info.hook_name is None:
                    # asked not to register for nothing
                    continue

                if not try_get_mark(method, 'register_if', True):
                    continue

                plugin_needs = list(
                    itertools.chain(
                        try_get_mark(method, 'plugin_needs', []),
                        global_needs,
                        registration_info.register_kwargs.get('needs', [])))

                plugin_provides = list(
                    itertools.chain(
                        try_get_mark(method, 'plugin_provides', []),
                        global_provides,
                        registration_info.register_kwargs.get('provides', [])))

                try:
                    if registration_info.expect_exists:
                        hook = gossip.get_hook(registration_info.hook_name)
                    else:
                        hook = gossip.hooks.get_or_create_hook(registration_info.hook_name)
                        if not hook.is_defined() and hook.group.is_strict():
                            raise LookupError()
                except LookupError:
                    unknown.append(registration_info.hook_name)
                    continue

                assert hook is not None
                register_no_op_hooks.discard(registration_info.hook_name)

                kwargs = registration_info.register_kwargs.copy()
                kwargs.update({
                    'needs': plugin_needs,
                    'provides': plugin_provides,
                    'token': self._get_token(plugin_name),
                })
                if registration_info.hook_name == 'slash.session_start':
                    has_session_start = True
                    kwargs['toggles_on'] = plugin.__toggles__['session']
                elif registration_info.hook_name == 'slash.session_end':
                    has_session_end = True
                    kwargs['toggles_off'] = plugin.__toggles__['session']

                returned.append((hook, method, kwargs))

        if has_session_end and not has_session_start:
            hook = gossip.get_hook('slash.session_start')
            returned.append((hook, lambda: None, {'toggles_on': plugin.__toggles__['session']}))
            register_no_op_hooks.discard(hook.full_name)

        for hook_name in register_no_op_hooks:
            hook = gossip.get_hook(hook_name)
            hook.register_no_op(provides=global_provides, token=self._get_token(plugin_name))

        if unknown:
            raise IncompatiblePlugin("Unknown hooks: {}".format(", ".join(unknown)))
        return returned


manager = PluginManager()



from __future__ import absolute_import

import collections

import logbook
import warnings

from . import hooks
from .utils.warning_capture import warning_callback_context
from .ctx import context
from contextlib import contextmanager


_native_logger = logbook.Logger('slash.native_warnings')


class LogbookWarning(UserWarning):
    pass

class SessionWarnings(object):
    """
    Holds all warnings emitted during the session
    """
    def __init__(self):
        super(SessionWarnings, self).__init__()
        self.warnings = []

    @contextmanager
    def capture_context(self):
        warnings.simplefilter('always')
        warnings.filterwarnings('ignore', category=ImportWarning)

        with warning_callback_context(self._capture_native_warning):
            yield

    def _capture_native_warning(self, message, category, filename, lineno, file=None, line=None): # pylint: disable=unused-argument
        warning = RecordedWarning.from_native_warning(message, category, filename, lineno)
        for ignored_warning in _ignored_warnings:
            if ignored_warning.matches(warning):
                return
        self.add(warning)
        if not issubclass(category, LogbookWarning):
            _native_logger.warning('{filename}:{lineno}: {warning!r}', filename=filename, lineno=lineno, warning=warning)

    def add(self, warning):
        hooks.warning_added(warning=warning) # pylint: disable=no-member
        self.warnings.append(warning)

    def __iter__(self):
        "Iterates through stored warnings"
        return iter(self.warnings)

    def __len__(self):
        return len(self.warnings)

    def __nonzero__(self):
        return bool(self.warnings)

    __bool__ = __nonzero__


class WarnHandler(logbook.Handler, logbook.StringFormatterHandlerMixin):
    """
    Like a stream handler but keeps the values in memory.
    This logger provides some ways to store warnings to log again at the end of the session.
    """
    default_format_string = (u'[{record.time:%Y-%m-%d %H:%M}] '
      '{record.level_name}: {record.extra[source]}: {record.message}')
    def __init__(self, session_warnings, format_string=None, filter=None, bubble=True):
        logbook.Handler.__init__(self, logbook.WARNING, filter, bubble)
        logbook.StringFormatterHandlerMixin.__init__(self, format_string)
        self.session_warnings = session_warnings

    def should_handle(self, record):
        """Returns `True` if this record is a warning """
        if record.channel == _native_logger.name:
            return False
        return record.level == self.level

    def emit(self, record):
        warnings.warn_explicit(message=record.message, category=LogbookWarning, filename=record.filename,
                               lineno=record.lineno, module=record.module)


WarningKey = collections.namedtuple("WarningKey", ("filename", "lineno"))

class RecordedWarning(object):

    def __init__(self, details, message, category=None):
        super(RecordedWarning, self).__init__()
        self.details = details
        self.details['session_id'] = context.session_id
        self.details['test_id'] = context.test_id
        self.details.setdefault('func_name', None)
        self.key = WarningKey(filename=self.details['filename'], lineno=self.details['lineno'])
        self.category = category
        self._repr = message

    @classmethod
    def from_log_record(cls, record, handler):
        details = record.to_dict()
        return cls(details, handler.format(record))


    @classmethod
    def from_native_warning(cls, message, category, filename, lineno):
        if isinstance(message, Warning):
            message = message.args[0]

        return cls({
            'message': message,
            'type': category.__name__,
            'filename': filename,
            'lineno': lineno,
            }, message=message, category=category)

    @property
    def message(self):
        return self.details.get('message')

    @property
    def lineno(self):
        return self.details.get('lineno')

    @property
    def filename(self):
        return self.details.get('filename')


    def to_dict(self):
        return self.details.copy()

    def __repr__(self):
        return self._repr

class _IgnoredWarning(object):

    def __init__(self, category, filename, lineno, message):
        self.category = category
        self.filename = filename
        self.lineno = lineno
        self.message = message

    def matches(self, warning):
        if self.category is not None and issubclass(warning.category, self.category):
            return True

        if self.filename is not None and warning.filename == self.filename:
            return True

        if self.lineno is not None and warning.lineno == self.lineno:
            return True

        if self.message is not None:
            if hasattr(self.message, 'match'):
                if self.message.match(warning.message):
                    return True
            elif self.message == warning.message:
                return True
        return False

_ignored_warnings = []


def ignore_warnings(category=None, message=None, filename=None, lineno=None):
    """
    Ignores warnings of specific origin (category/filename/lineno/message) during the session. Unlike
    Python's default ``warnings.filterwarnings``, the parameters are matched only if specified (not defaulting to "match all"). Message can also be a
    regular expression object compiled with ``re.compile``.

        slash.ignore_warnings(category=CustomWarningCategory)
    """
    _ignored_warnings.append(_IgnoredWarning(category=category, filename=filename, lineno=lineno, message=message))


def clear_ignored_warnings():
    del _ignored_warnings[:]



import threading
import logbook
import os
import pickle
import time
from six.moves import xmlrpc_client
from .server import NO_MORE_TESTS, PROTOCOL_ERROR, WAITING_FOR_CLIENTS
from ..exceptions import INTERRUPTION_EXCEPTIONS
from ..hooks import register
from ..ctx import context
from ..runner import run_tests
from ..conf import config
_logger = logbook.Logger(__name__)

class Worker(object):
    def __init__(self, client_id, session_id):
        super(Worker, self).__init__()
        self.client_id = client_id
        self.session_id = session_id
        self.server_addr = 'http://{}:{}'.format(config.root.parallel.server_addr, config.root.parallel.server_port)

    def keep_alive(self, stop_event):
        proxy = xmlrpc_client.ServerProxy(self.server_addr)
        while not stop_event.is_set():
            proxy.keep_alive(self.client_id)
            stop_event.wait(1)

    def warning_added(self, warning):
        try:
            warning = pickle.dumps(warning)
            self.client.report_warning(self.client_id, warning)
        except (pickle.PicklingError, TypeError):
            _logger.error("Failed to pickle warning. Message: {}, File: {}, Line: {}",
                          warning.message, warning.filename, warning.lineno)

    def error_added(self, error, result):
        if result.is_global_result():
            self.client.report_session_error(self.client_id, error.message)

    def write_to_error_file(self, msg):
        try:
            file_name = "{}-{}.log".format(config.root.parallel.worker_error_file, self.client_id)
            worker_error_file = os.path.join(config.root.parallel.workers_error_dir, file_name)
            with open(worker_error_file, 'a') as error_file:
                error_file.write(msg)
        except OSError as err:
            _logger.error("Failed to write to worker error file, error: {}", err)

    def connect_to_server(self):
        try:
            self.client = xmlrpc_client.ServerProxy(self.server_addr, allow_none=True)
            self.client.connect(self.client_id, os.getpid())
        except OSError as err:
            self.write_to_error_file("Failed to connect to server, error: {}".format(str(err)))

    def start_execution(self, app, collected_tests):
        if os.getpid() != os.getpgid(0):
            os.setsid()
        register(self.warning_added)
        register(self.error_added)
        collection = [(test.__slash__.file_path, test.__slash__.function_name, test.__slash__.variation.id) for test in collected_tests]

        if not self.client.validate_collection(self.client_id, collection):
            _logger.error("Collections of worker id {} and master don't match, worker terminates", self.client_id)
            self.client.disconnect(self.client_id)
            return

        stop_event = threading.Event()
        watchdog_thread = threading.Thread(target=self.keep_alive, args=(stop_event,))
        watchdog_thread.setDaemon(True)
        watchdog_thread.start()
        should_stop = False
        with app.session.get_started_context():
            try:
                while not should_stop:
                    test_index = self.client.get_test(self.client_id)
                    if test_index == WAITING_FOR_CLIENTS:
                        _logger.debug("Worker_id {} recieved waiting_for_clients, sleeping", self.client_id)
                        time.sleep(0.05)
                    elif test_index == PROTOCOL_ERROR:
                        _logger.error("Worker_id {} recieved protocol error message, terminating", self.client_id)
                        break
                    elif test_index == NO_MORE_TESTS:
                        _logger.debug("Got NO_MORE_TESTS, Client {} disconnecting", self.client_id)
                        break
                    else:
                        test = collected_tests[test_index]
                        context.session.current_parallel_test_index = test_index
                        run_tests([test])
                        result = context.session.results[test]
                        _logger.debug("Client {} finished test, sending results", self.client_id)
                        ret = self.client.finished_test(self.client_id, result.serialize())
                        if ret == PROTOCOL_ERROR:
                            _logger.error("Worker_id {} recieved protocol error message, terminating", self.client_id)
                            should_stop = True
            except INTERRUPTION_EXCEPTIONS:
                self.write_to_error_file("Worker interrupted while executing test")
                _logger.error("Worker interrupted while executing test")
                raise
            except Exception as err:
                self.write_to_error_file(str(err))
                raise
            else:
                context.session.mark_complete()
                context.session.initiate_cleanup()
                stop_event.set()
                watchdog_thread.join()
                self.client.disconnect(self.client_id)
            finally:
                if not stop_event.is_set():
                    stop_event.set()
                    watchdog_thread.join()



import pytest
import slash

import itertools
from .utils import run_tests_in_session


@pytest.mark.parametrize('on_fixture', [True, False])
def test_stable_ordering(on_fixture):
    """Make sure that cartesian products in parametrizations are stable, even when changing the order of the parametrization decorators
    """
    params = [('a', [1, 2, 3]), ('b', [True, False]), ('c', [4, 5, 6])]
    permutations = list(itertools.permutations(params))

    def parametrize(permutation):
        def decorator(thing):
            for param_name, values in permutation:
                thing = slash.parametrize(param_name, values)(thing)
            return thing
        return decorator

    def generate_session_and_test(permutation):

        s = slash.Session()

        if on_fixture:

            @parametrize(permutation)
            def some_fixture(a, b, c):
                return {'a': a, 'b': b, 'c': c}

            def test_something(some_fixture):
                slash.context.result.data['params'] = some_fixture
        else:
            @parametrize(permutation)
            def test_something(a, b, c):
                slash.context.result.data['params'] = {'a': a, 'b': b, 'c': c}


        if on_fixture:
            s.fixture_store.add_fixture(slash.fixture(some_fixture))

        s.fixture_store.resolve()

        return s, test_something

    results_by_permutation = []
    for p in permutations:
        session, test = generate_session_and_test(p)
        with session, session.get_started_context():
            run_tests_in_session(test, session=session)
        assert session.results.is_success(), 'run failed'
        results_by_permutation.append(session.results)

    param_values = {}
    for results in results_by_permutation:
        for index, result in enumerate(results):
            for param_name, param_value in result.data['params'].items():
                param_values.setdefault((param_name, index), []).append(param_value)

    for (param_name, index), values in param_values.items():
        assert len(set(values)) == 1, '{}#{} is not stable: {}'.format(param_name, index, values)



# pylint: disable=redefined-outer-name
import pytest

import slash
from slash._compat import cStringIO
from slash.frontend.slash_run import slash_run


def test_errors_during_initialization_hoook(suite, init_hook):

    @init_hook.register
    def callback():  # pylint: disable=unused-variable
        raise slash.exceptions.SlashException('some error')

    exit_code, output = _console_run(suite)
    assert 'some error' in output
    assert exit_code != 0


def test_slashrc_errors(suite):
    @suite.slashrc.include
    def __code__():  # pylint: disable=unused-variable
        1 / 0  # pylint: disable=pointless-statement

    exit_code, output = _console_run(suite)
    assert exit_code != 0
    output = output.lower()
    assert 'unexpected error' in output
    assert 'division' in output
    assert 'zero' in output


@pytest.fixture(params=[
    slash.hooks.session_start,  # pylint: disable=no-member
    slash.hooks.configure,  # pylint: disable=no-member
])
def init_hook(request):
    return request.param


def _console_run(suite):
    suite.disable_debug_info()
    path = suite.commit()
    stream = cStringIO()
    exit_code = slash_run([path], report_stream=stream, working_directory=path)
    return exit_code, stream.getvalue()



# pylint: disable=unused-argument, unused-variable
from uuid import uuid4

import slash

from .conftest import Checkpoint
from .utils import make_runnable_tests


def test_yield_fixture(yield_fixture_decorator):
    iterations = [uuid4() for i in range(3)]
    start_checkpoints, end_checkpoints = [
        {iteration: Checkpoint() for iteration in iterations}
        for i in range(2)]
    value = uuid4()
    inner_fixture_value = uuid4()

    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.fixture
        def other_fixture():
            return inner_fixture_value


        @s.fixture_store.add_fixture
        @slash.parametrize('iteration', list(iterations))
        @yield_fixture_decorator
        def fixture(iteration, other_fixture):
            assert other_fixture == inner_fixture_value
            start_checkpoints[iteration]()
            yield value
            end_checkpoints[iteration]()

        def test_something(fixture):
            assert fixture == value

        s.fixture_store.resolve()

        with s.get_started_context():
            slash.runner.run_tests(make_runnable_tests(test_something))
        assert s.results.is_success(allow_skips=False)


def test_yield_fixture_with_this_argument(yield_fixture_decorator):
    iterations = [uuid4() for i in range(3)]
    value = uuid4()

    with slash.Session() as s:


        @s.fixture_store.add_fixture
        @slash.parametrize('iteration', list(iterations))
        @yield_fixture_decorator
        def fixture(this, iteration):
            yield value

        def test_something(fixture):
            assert fixture == value

        s.fixture_store.resolve()

        with s.get_started_context():
            slash.runner.run_tests(make_runnable_tests(test_something))


def test_yield_fixture_with_scope_argument(yield_fixture_decorator):
    value = str(uuid4())

    with slash.Session() as s:


        @s.fixture_store.add_fixture
        @yield_fixture_decorator(scope='session')
        def fixture():
            yield value

        def test_something(fixture):
            assert fixture == value

        s.fixture_store.resolve()

        with s.get_started_context():
            slash.runner.run_tests(make_runnable_tests(test_something))



from slash.utils.pattern_matching import Matcher
from slash.core.tagging import NO_TAGS, Tags


def test_string_patterns(suite, config_override):
    selected_test = suite[len(suite) // 2]

    config_override("run.filter_strings", [selected_test.id])

    for i, test in enumerate(suite):  # pylint: disable=unused-variable
        if test is not selected_test:
            test.expect_deselect()

    suite.run()


def test_string_patterns_and(suite, config_override):
    selected_test = suite[len(suite) // 2]
    other_test = suite[-1]

    config_override("run.filter_strings", [selected_test.id, other_test.id])

    for i, test in enumerate(suite):  # pylint: disable=unused-variable
        test.expect_deselect()

    result = suite.run(expect_session_errors=True).session.results.global_result
    assert len(result.get_errors()) == 1
    assert 'no tests could be collected' in str(result.get_errors()[0]).lower()


def test_matcher():
    for pattern, matching, not_matching in [
            ('xy', ['blapxy'], ['blap']),
            ('not x', ['blap', 'yyy'], ['blapx', 'xy']),
            ('x and y', ['xy', 'yx', 'x kjfldkjfd y'], ['x', 'y', 'xx']),
            ('(x and some.string) and not z', ['xysome.string', 'some.stringyx'], ['z', 'xysome.stringz']),
            ('exact(x=y)(y=z)', ['exact(x=y)(y=z)'], ['exlact(x=y)(y=z)']),
    ]:

        matcher = Matcher(pattern)
        for string in matching:
            assert matcher.matches(FakeMetadata(string))
        for string in not_matching:
            assert not matcher.matches(FakeMetadata(string))


def test_test_name_matcher():
    test_name = '/path/to/testfile.py:TestName'
    m = FakeMetadata(test_name)
    assert not Matcher('not {}'.format(test_name)).matches(m)
    assert Matcher(test_name).matches(m)


def test_matches_tag():
    assert Matcher('bla').matches(FakeMetadata('something', {'bla': 2}))
    assert not Matcher('bla').matches(FakeMetadata('something', {'bloop': 2}))
    assert Matcher('substring').matches(FakeMetadata('something', {'xxxxsubstringxxx': 2}))


def test_matches_tag_exclusively():
    assert Matcher('tag:bla').matches(FakeMetadata('something', {'bla': 2}))
    assert not Matcher('tag:bla').matches(FakeMetadata('bla', {'bloop': 2}))


def test_matches_values():
    assert Matcher('tag:bla=2').matches(FakeMetadata('something', {'bla': 2}))
    assert not Matcher('tag:bla=2').matches(FakeMetadata('something', {'bla': 3}))
    assert Matcher('tag:bla=hello').matches(FakeMetadata('something', {'bla': 'hello'}))
    assert not Matcher('tag:bla=bye').matches(FakeMetadata('something', {'bla': 'hello'}))


class FakeMetadata(object):

    def __init__(self, address, tags=NO_TAGS):
        super(FakeMetadata, self).__init__()
        if isinstance(tags, dict):
            tags = Tags(tags)
        self.address = address
        self.tags = tags



# pylint: disable=redefined-outer-name
import os

import pytest

from xml.etree import ElementTree


def test_xunit_plugin(results, xunit_filename): # pylint: disable=unused-argument
    assert os.path.exists(xunit_filename), 'xunit file not created'

    validate_xml(xunit_filename)


def test_session_errors(suite, xunit_filename):
    # pylint: disable=unused-argument

    @suite.slashconf.append_body
    def __code__():  # pylint: disable=unused-variable
        1/0  # pylint: disable=pointless-statement

    for test in suite:
        # tests are not going to even be loaded
        test.expect_deselect()

    summary = suite.run(expect_session_errors=True)

    assert 'ZeroDivision' in summary.get_console_output()



def test_xunit_plugin_test_details(suite, suite_test, xunit_filename, details):
    for key, value in details.items():
        suite_test.append_line('slash.context.result.set_test_detail({!r}, {!r})'.format(key, value))

    suite.run()
    testcase_xml = _get_testcase_xml(suite_test, xunit_filename)

    saved_details = dict((d.attrib['name'], d.attrib['value']) for d in testcase_xml.findall('detail'))
    assert saved_details


@pytest.mark.parametrize('errtype', ['error', 'failure'])
def test_xunit_plugin_add_failure_error(suite, suite_test, xunit_filename, errtype):
    num_errors = 3
    for _ in range(num_errors):
        suite_test.append_line('slash.add_{}("some message")'.format(errtype))
    if errtype == 'error':
        suite_test.expect_error()
    else:
        suite_test.expect_failure()

    suite.run()
    testcase_xml = _get_testcase_xml(suite_test, xunit_filename)
    errors = testcase_xml.findall(errtype)
    assert len(errors) == num_errors
    for error in errors:
        assert error.attrib['message'] == 'some message'
        assert error.attrib['type'] == errtype
    assert errors



def _get_testcase_xml(suite_test, filename):
    with open(filename) as f:
        xml = ElementTree.parse(f)
    match = [testcase for testcase in list(xml.getroot()) if testcase.get('name').split('_')[-1] == suite_test.id]
    assert len(match) == 1
    return match[0]


@pytest.fixture
def details():
    return {'detail1': 'value1', 'detail2': 'value2'}


@pytest.fixture  # pylint: disable=unused-argument
def results(suite, suite_test, test_event, xunit_filename): # pylint: disable=unused-argument
    test_event(suite_test)
    summary = suite.run()
    assert 'Traceback' not in summary.get_console_output()

@pytest.fixture(params=['normal', 'skip_decorator_without_reason', 'skip_without_reason', 'skip_with_reason',
                        'error', 'failure', 'add_error', 'add_failure'])
def test_event(request):
    flavor = request.param

    def func(test):
        if flavor != 'normal':
            if flavor == 'skip_without_reason':
                test.when_run.skip(with_reason=False)
            elif flavor == 'skip_decorator_without_reason':
                test.when_run.skip(with_reason=False, decorator=True)
            elif flavor == 'skip_with_reason':
                test.when_run.skip(with_reason=True)
            elif flavor == 'error':
                test.when_run.error()
            elif flavor == 'failure':
                test.when_run.fail()
            elif flavor == 'add_error':
                test.append_line('slash.add_error("error")')
                test.expect_error()
            elif flavor == 'add_failure':
                test.append_line('slash.add_failure("failure")')
                test.expect_failure()
            else:
                raise NotImplementedError()  # pragma: no cover
    return func


@pytest.mark.parametrize('action', ['skipped', 'error', 'failure', 'success'])
def test_xunit_parallel_suite(parallel_suite, xunit_filename, action):
    test = parallel_suite[3]
    if action == 'skipped':
        test.when_run.skip()
    elif action == 'failure':
        test.when_run.fail()
    elif action == 'error':
        test.when_run.raise_exception()
    else:
        assert action == 'success', 'Unsupported action'
    parallel_suite.run()

    root = validate_xml(xunit_filename, suite=parallel_suite)
    if action != 'success':
        for child in root:
            if len(child) > 0: # pylint: disable=len-as-condition
                [element] = child
                break
        else:
            assert False, 'no child matching test found'
        assert element.tag == action



def validate_xml(xml_filename, suite=None):
    with open(xml_filename) as f:
        etree = ElementTree.parse(f)

    root = etree.getroot()
    assert root.tag == 'testsuite'

    _validate_counters(root)

    if suite is not None:
        assert len(root) == len(suite)

    for child in list(root):
        assert child.tag == 'testcase'
        assert child.get('name')
        assert child.get('time')

        for subchild in child.getchildren():
            assert subchild.tag in ['skipped', 'error', 'failure']
    return root

def _validate_counters(element):
    for number in ['errors', 'failures', 'skipped']:
        _ = int(element.get(number))



# pylint: disable=unused-variable,redefined-outer-name
import collections
import warnings
from contextlib import contextmanager
from uuid import uuid4
from vintage import deprecated

import logbook
import pytest
import inspect

import slash

from .utils import run_tests_assert_success, without_pyc


@pytest.mark.parametrize('reprify', [repr, str])
def test_str_repr(warning, reprify):
    assert 'this is a warning' in reprify(warning)


def test_location(warning):
    assert warning.details['filename'] == without_pyc(__file__)


def test_to_dict(warning):
    assert isinstance(warning.to_dict(), dict)


def test_warning_added_hook(suite, suite_test):

    captured = []

    @slash.hooks.register
    def warning_added(warning):
        captured.append(warning)

    suite_test.append_line('slash.logger.warning("message {}", "here")')
    suite.run()
    assert captured
    [w] = captured # pylint: disable=unbalanced-tuple-unpacking
    assert w.message == 'message here'
    assert isinstance(w.lineno, int)
    assert isinstance(w.filename, str)
    assert w.lineno
    assert w.filename
    assert w.filename.rsplit('/', 1)[-1] == suite_test.file.get_relative_path()
    warning_type = w.details['type']
    assert isinstance(warning_type, str)
    assert warning_type == 'LogbookWarning'


def _get_current_line_info():
    frame = inspect.currentframe().f_back
    return frame.f_code.co_filename, frame.f_lineno


def test_native_warnings(message):

    def test_example():
        with logbook.TestHandler() as handler:
            filename, line_no = _get_current_line_info()
            warnings.warn(message)
        assert len(handler.records) == 1
        rec_message = handler.records[0].message
        assert rec_message.endswith(message)
        assert filename in rec_message
        assert str(line_no + 1) in rec_message

    s = run_tests_assert_success(test_example)

    assert len(s.warnings) == 1
    [w] = s.warnings
    assert w.message == message


def test_deprecation(message):
    @deprecated(message=message)
    def deprecated_func():
        pass
    with capturing_native_warnings() as handlers:
        deprecated_func()
    assert len(handlers.native_warnings) == 1
    native_warning = handlers.native_warnings[0]
    warning_message = native_warning.message.args[0]
    assert message in warning_message


@pytest.fixture
def message():
    return 'some message here {}'.format(uuid4())


@pytest.fixture
def warning():
    class SampleTest(slash.Test):

        def test(self):
            slash.logger.warning("this is a warning. Param: {0}", 1)

    session = run_tests_assert_success(SampleTest)

    assert len(session.warnings) == 1

    [warning] = session.warnings
    return warning


@contextmanager
def capturing_native_warnings():
    with logbook.TestHandler() as log_handler:
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter('always')
            handlers = collections.namedtuple('Handlers', ('log', 'native_warnings'))(log_handler, recorded)
            yield handlers



import datetime
import numbers
import os
import sys
from contextlib import contextmanager, closing

import logbook  # pylint: disable=F0401
import logbook.more
from vintage import warn_deprecation

from . import context
from ._compat import ExitStack
from .conf import config
from .utils.path import ensure_containing_directory
from .warnings import WarnHandler
from .exceptions import InvalidConfiguraion
from .exception_handling import handling_exceptions
from . import hooks

_logger = logbook.Logger(__name__)

_custom_colors = {}
filtered_channels = {'slash.runner', 'slash.loader', 'slash.core.cleanup_manager', 'slash.core.scope_manager', \
                      'slash.exception_handling', 'slash.core.fixtures.fixture_store'}

class _NormalizedObject(object):
    def __init__(self, obj):
        self._obj = obj

    def __getattr__(self, name):
        obj = getattr(self._obj, name)
        return obj if isinstance(obj, numbers.Number) else _NormalizedObject(obj)

    @staticmethod
    def _escape(s):
        return s.replace('\\', '_').replace('/', '_')

    def __str__(self):
        return self._escape(str(self._obj))

    def __repr__(self):
        return self._escape(repr(self._obj))


def set_log_color(logger_name, level, color):
    """Sets the color displayed in the console, according to the logger name and level
    """
    _custom_colors[logger_name, level] = color

class ColorizedHandlerMixin(logbook.more.ColorizingStreamHandlerMixin):

    def get_color(self, record):
        returned = _custom_colors.get((record.channel, record.level))
        if returned is not None:
            return returned

        if record.level >= logbook.ERROR:
            return 'red'
        elif record.level >= logbook.WARNING:
            return 'yellow'
        elif record.level >= logbook.NOTICE:
            return 'white'
        return None # default

class ColorizedFileHandler(ColorizedHandlerMixin, logbook.FileHandler):

    def should_colorize(self, record):
        return True


class ConsoleHandler(ColorizedHandlerMixin, logbook.StreamHandler):

    MAX_LINE_LENGTH = 160

    default_format_string = '[{record.time:%Y-%m-%d %H:%M:%S}] {record.message}'

    def __init__(self, **kw):
        stream = kw.pop('stream', sys.stderr)
        logbook.StreamHandler.__init__(self, stream=stream, **kw)
        self._truncate_lines = config.root.log.truncate_console_lines
        self._truncate_errors = config.root.log.truncate_console_errors
        if config.root.log.color_console is not None:
            if config.root.log.color_console:
                self.force_color()
            else:
                self.forbid_color()

    def format(self, record):
        orig_message = record.message
        should_truncate = self._truncate_errors or record.level < logbook.ERROR
        if self._truncate_lines and len(str(orig_message)) > self.MAX_LINE_LENGTH and should_truncate:
            record.message = "\n".join(self._truncate(line) for line in orig_message.splitlines())
        try:
            returned = super(ConsoleHandler, self).format(record)
        finally:
            # we back up the original message to avoid propagating truncated lines into the file log
            record.message = orig_message
        return returned

    def _truncate(self, line):
        if len(line) > self.MAX_LINE_LENGTH:
            line = line[:self.MAX_LINE_LENGTH - 3] + "..."
        return line

    def emit(self, record):
        reporter = None if context.session is None else context.session.reporter
        if reporter is not None:
            reporter.notify_before_console_output()
        returned = super(ConsoleHandler, self).emit(record)
        if reporter is not None:
            reporter.notify_after_console_output()
        return returned

def _slash_logs_filter(record, handler): # pylint: disable=unused-argument
    return record.extra.get('filter_bypass') or \
           record.channel not in filtered_channels or \
           record.level >= config.root.log.core_log_level

class SessionLogging(object):
    """
    A context creator for logging within a session and its tests
    """
    def __init__(self, session, console_stream=None):
        super(SessionLogging, self).__init__()
        if console_stream is None:
            console_stream = sys.stderr
        self.session = session
        self.warnings_handler = WarnHandler(session.warnings)
        self.console_handler = ConsoleHandler(bubble=True, level=config.root.log.console_level, stream=console_stream, filter=_slash_logs_filter)
        #: contains the path for the session logs
        self.session_log_path = None
        self.session_log_handler = None
        #: contains the path for the current test logs
        self.test_log_path = None
        self._set_formatting(self.console_handler, config.root.log.console_format or config.root.log.format)
        self._log_path_to_handler = {}

    def get_active_log_paths(self):
        return [log_path for log_path, handler in self._log_path_to_handler.items() if handler is not None]

    @contextmanager
    def get_test_logging_context(self, result):
        with self._get_file_logging_context(config.root.log.subpath, config.root.log.last_test_symlink) as (_, path):
            self.test_log_path = path
            result.set_log_path(path)
            try:
                yield path
            finally:
                self._create_last_failed_symlink_if_needed(result)

    def _create_last_failed_symlink_if_needed(self, result):
        assert result
        if result.is_error() or result.is_failure():
            self._try_create_symlink(result.get_log_path(), config.root.log.last_failed_symlink)

    @contextmanager
    def get_session_logging_context(self):
        assert self.session_log_handler is None
        with self._get_file_logging_context(
            config.root.log.session_subpath, config.root.log.last_session_symlink) as (handler, path):
            self.session_log_handler = handler
            self.session_log_path = path
            self.session.results.global_result.set_log_path(path)
            if config.root.log.last_session_dir_symlink is not None and self.session_log_path is not None:
                self._try_create_symlink(os.path.dirname(self.session_log_path), config.root.log.last_session_dir_symlink)
            yield path

    @contextmanager
    def _get_file_logging_context(self, filename_template, symlink):
        with ExitStack() as stack:
            handler = stack.enter_context(self._log_file_handler_context(filename_template, symlink, \
                                                                         use_compression=config.root.log.compression.enabled))
            stack.enter_context(handler.applicationbound())
            if config.root.log.compression.enabled and config.root.log.compression.use_rotating_raw_file:
                rotating_handler = stack.enter_context(self._log_file_handler_context(filename_template, symlink, bubble=True, use_rotation=True))
                stack.enter_context(rotating_handler.applicationbound())

            stack.enter_context(self.console_handler.applicationbound())
            stack.enter_context(self.warnings_handler.applicationbound())
            error_handler = stack.enter_context(self._get_error_logging_context())
            stack.enter_context(error_handler.applicationbound())
            stack.enter_context(self._get_silenced_logs_context())
            if config.root.log.unittest_mode:
                stack.enter_context(logbook.StreamHandler(sys.stderr, bubble=True, level=logbook.TRACE))
            for extra_handler in _extra_handlers:
                stack.enter_context(extra_handler.applicationbound())
            if config.root.log.unified_session_log and self.session_log_handler is not None:
                stack.enter_context(_make_bubbling_handler(self.session_log_handler))

            path = handler.stream.name if isinstance(handler, logbook.FileHandler) else None
            yield handler, path


    def _should_delete_log(self, result):
        return (not config.root.log.cleanup.keep_failed) or \
               (not result.is_global_result() and result.is_success(allow_skips=True)) or \
               (result.is_global_result() and self.session.results.is_success(allow_skips=True))

    @contextmanager
    def _get_error_logging_context(self):
        with ExitStack() as stack:
            path = config.root.log.errors_subpath
            if path:
                warn_deprecation('log.errors_subpath configuration is deprecated since 1.5.0. '
                                 'Please use log.highlights_subpath instead')
            else:
                path = config.root.log.highlights_subpath
            def _error_added_filter(record, handler): # pylint: disable=unused-argument
                return record.extra.get('highlight')

            handler = stack.enter_context(self._log_file_handler_context(path, symlink=None, bubble=True, filter=_error_added_filter))
            log_path = handler.stream.name if isinstance(handler, logbook.FileHandler) else None
            if log_path and self.session.results.current is self.session.results.global_result:
                self.session.results.global_result.add_extra_log_path(log_path)
            yield handler

    def _get_silenced_logs_context(self):
        if not config.root.log.silence_loggers:
            return ExitStack()
        return SilencedLoggersHandler(config.root.log.silence_loggers).applicationbound()

    def _get_log_file_path(self, subpath, use_compression):
        log_path = self._normalize_path(os.path.join(config.root.log.root, _format_log_path(subpath)))
        if use_compression:
            if config.root.log.compression.algorithm == "gzip":
                log_path += ".gz"
            elif config.root.log.compression.algorithm == "brotli":
                log_path += ".br"
            else:
                raise InvalidConfiguraion("Unsupported compression method: {}".format(config.root.log.compression.algorithm))
        return log_path

    def _create_log_file_handler(self, log_path, bubble=False, filter=_slash_logs_filter, use_compression=False, use_rotation=False):
        kwargs = {"bubble": bubble, "filter": filter}
        if use_compression:
            if config.root.log.compression.algorithm == "gzip":
                handler_class = logbook.GZIPCompressionHandler
            elif config.root.log.compression.algorithm == "brotli":
                handler_class = logbook.BrotliCompressionHandler
        elif use_rotation:
            kwargs.update({"max_size": 4*1024**2, "backup_count": 1})
            handler_class = logbook.RotatingFileHandler
        elif config.root.log.colorize:
            handler_class = ColorizedFileHandler
        else:
            handler_class = logbook.FileHandler
        return handler_class(log_path, **kwargs)

    @contextmanager
    def _log_file_handler_context(self, subpath, symlink, bubble=False, filter=_slash_logs_filter, use_compression=False, use_rotation=False):
        if subpath is None or config.root.log.root is None:
            yield NoopHandler() if bubble else logbook.NullHandler(filter=filter)
        else:
            log_path = self._get_log_file_path(subpath, use_compression)
            handler = self._log_path_to_handler.get(log_path, None)
            if handler is not None:
                yield handler
            else:
                result = context.result
                ensure_containing_directory(log_path)
                if symlink:
                    self._try_create_symlink(log_path, symlink)
                with closing(self._create_log_file_handler(log_path, bubble=bubble, use_compression=use_compression, \
                                                           use_rotation=use_rotation, filter=filter)) as handler:
                    self._log_path_to_handler[log_path] = handler
                    self._set_formatting(handler, config.root.log.format)
                    yield handler
                self._log_path_to_handler[log_path] = None
                hooks.log_file_closed(path=log_path, result=result)  # pylint: disable=no-member
                if config.root.log.cleanup.enabled and self._should_delete_log(result):
                    with handling_exceptions(swallow=True):
                        os.remove(log_path)
                        dir_path = os.path.dirname(log_path)
                        if not os.listdir(dir_path) and dir_path != self._normalize_path(config.root.log.root):
                            os.rmdir(dir_path)


    def _normalize_path(self, p):
        return os.path.expanduser(p)

    def create_worker_symlink(self, worker_name, worker_session_id):
        if config.root.log.root is None:
            return
        symlink = os.path.join(self.session.id, worker_name)
        worker_dir = os.path.join(self._normalize_path(config.root.log.root), worker_session_id)
        self._try_create_symlink(worker_dir, symlink)

    def _try_create_symlink(self, path, symlink):
        if symlink is None or config.root.log.root is None or config.root.parallel.worker_id is not None:
            return

        symlink = self._normalize_path(symlink)

        if not os.path.isabs(symlink):
            symlink = os.path.join(self._normalize_path(config.root.log.root), symlink)

        try:
            ensure_containing_directory(symlink)

            if os.path.exists(symlink) or os.path.islink(symlink):
                os.unlink(symlink)
            os.symlink(path, symlink)

        except Exception:  # pylint: disable=broad-except
            _logger.debug("Failed to create symlink {0} --> {1}", path, symlink, exc_info=True)

    def _set_formatting(self, handler, fmt):
        if config.root.log.localtime:
            logbook.set_datetime_format("local")
        if fmt is not None:
            handler.format_string = fmt

class SilencedLoggersHandler(logbook.Handler):
    def __init__(self, silence_logger_names):
        super(SilencedLoggersHandler, self).__init__(bubble=False)
        self._silenced_names = set(silence_logger_names)
    def should_handle(self, record):
        return record.channel in self._silenced_names

class NoopHandler(object):

    # Logbook's NullHandler does not bubble by default. This is dummy handler that
    # does not interfere with the stack at all
    def applicationbound(self):
        return ExitStack()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        pass

def add_log_handler(handler):
    """
    Adds a log handler to be entered for sessions and for tests
    """
    _extra_handlers.append(handler)

def remove_all_extra_handlers():
    del _extra_handlers[:]

_extra_handlers = []

class VERBOSITIES(object):

    DEBUG = logbook.DEBUG
    INFO = logbook.INFO
    NOTICE = logbook.NOTICE
    WARNING = logbook.WARNING
    ERROR = logbook.ERROR
    CRITICAL = logbook.CRITICAL


def _make_bubbling_handler(handler):
    return _BubblingWrapper(handler)

class _BubblingWrapper(logbook.Handler):

    def __init__(self, handler):
        super(_BubblingWrapper, self).__init__(bubble=True)
        self._handler = handler
        self.should_handle = self._handler.should_handle
        self.handle = self._handler.handle


def _format_log_path(p):
    return p.format(context=_NormalizedObject(context), timestamp=datetime.datetime.now(),
                    project_name=config.root.run.project_name)


class RetainedLogHandler(logbook.TestHandler):
    """A logbook handler that retains the emitted logs in order to
    flush them later to a handler.

    This is useful to keep logs that are emitted during session configuration phase, and not lose
    them from the session log
    """
    def __init__(self, *args, **kwargs):
        super(RetainedLogHandler, self).__init__(*args, **kwargs)
        self._enabled = True

    def emit(self, record):
        if self._enabled:
            return super(RetainedLogHandler, self).emit(record)

    def flush_to_handler(self, handler):
        for r in self.records:
            if handler.should_handle(r):
                handler.emit(r)
        del self.records[:]

    def disable(self):
        self._enabled = False



from uuid import uuid4

from .function import Function


class GeneratorFixture(Function):

    def __init__(self, suite, file, num_values=3):
        super(GeneratorFixture, self).__init__(suite)
        self.file = file
        self.values = [str(uuid4()) for _ in range(num_values)]

    def is_generator_fixture(self):
        return True

    def _write_decorators(self, code_formatter):
        code_formatter.writeln('@slash.generator_fixture')
        super(GeneratorFixture, self)._write_decorators(code_formatter)

    def _get_function_name(self):
        return 'fx_{}'.format(self.id)

    def _write_return(self, code_formatter):
        for value in self.values:
            code_formatter.writeln(
                'yield {!r}'.format(value))

    def _get_argument_strings(self):
        return []

    def __repr__(self):
        return '<Generator Fixture {}>'.format(self.name)



from __future__ import print_function

import functools
from contextlib import contextmanager

import slash

from .utils import run_tests_assert_success


def test_fixture_is_override():
    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.fixture
        def fixture1():
            pass

        outer = fixture1

        with s.fixture_store.new_namespace_context():

            @s.fixture_store.add_fixture
            @slash.fixture
            def fixture1():     # pylint: disable=function-redefined
                pass

            inner = fixture1

            s.fixture_store.resolve()

        outer = s.fixture_store.get_fixture_by_id(outer.__slash_fixture__.id)
        inner = s.fixture_store.get_fixture_by_id(inner.__slash_fixture__.id)

        assert outer != inner

        assert inner.is_override()
        assert not outer.is_override()


def test_fixture_namespaces(tmpdir):

    root_dir = tmpdir.join('tests')

    with _into(root_dir.join('slashconf.py')) as writeln:
        _write_fixture1(writeln)

    with _into(root_dir.join('test_something.py')) as writeln:
        _write_fixture1(writeln)

        writeln('def test_something(fixture1):')
        writeln('    pass')

    run_tests_assert_success([str(root_dir)])

def _write_fixture1(writeln):
    writeln('import slash')

    writeln('@slash.fixture')
    writeln('def fixture1():')
    writeln('    pass')



@contextmanager
def _into(f):
    with f.open('w', ensure=True) as f:
        yield functools.partial(print, file=f)



# pylint: disable=unused-variable, unused-argument
import functools

import pytest

import slash
from slash.exceptions import UnknownFixtures

from .utils import make_runnable_tests


def test_resolve_fixture_object():
    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.parametrize('x', [1, 2])
        @slash.fixture
        def fixture1(x):
            pass

        @s.fixture_store.add_fixture
        @slash.parametrize('x', [3, 4])
        @slash.fixture
        def fixture2(fixture1, x):
            pass

        def test_something(fixture2):
            pass

        s.fixture_store.resolve()

        with s.get_started_context():
            tests = make_runnable_tests(test_something)

    _resolve = functools.partial(s.fixture_store.resolve_name, start_point=tests[0])
    def _resolve_values(path):
        return [v.value for v in _resolve(path).values]

    # check simple resolutions
    assert _resolve('fixture2').info is fixture2.__slash_fixture__
    assert _resolve('fixture2.fixture1').info is fixture1.__slash_fixture__

    # check parameter resolution
    assert _resolve_values('fixture2.fixture1.x') == [1, 2]
    assert _resolve_values('fixture2.x') == [3, 4]

    for invalid_name in ['fixture2.x.y']:
        with pytest.raises(UnknownFixtures):
            _resolve(invalid_name)

def test_resolve_fixture_object_namespace_correctness():
    with slash.Session() as s:

        store = s.fixture_store

        @store.add_fixture
        @slash.fixture
        def global_fixture_1(dependency_fixture):
            pass

        @store.add_fixture
        @slash.fixture
        def dependency_fixture():
            pass

        expected = dependency_fixture

        store.push_namespace()

        @store.add_fixture
        @slash.fixture
        def local_fixture(global_fixture_1):
            pass

        @store.add_fixture
        @slash.fixture
        def dependency_fixture():         # pylint: disable=function-redefined
            pass

        def test_something(local_fixture):
            pass

        store.resolve()

        with s.get_started_context():
            tests = make_runnable_tests(test_something)

        test = tests[0]
        resolved = store.resolve_name('local_fixture.global_fixture_1.dependency_fixture', test, namespace=test.get_fixture_namespace()).info
        assert resolved is expected.__slash_fixture__


def test_invalid_name():
    with slash.Session() as s:
        with pytest.raises(UnknownFixtures):
            s.fixture_store.resolve_name('', start_point=object())



import datetime
import itertools
import socket
import sys
from ..interface import PluginInterface
from ...ctx import context
from ...conf import config
from ...utils.traceback_utils import get_traceback_string
from ...utils.conf_utils import Cmdline, Doc
from slash import config as slash_config
from slash import context
from xml.etree.ElementTree import (
    tostring as xml_to_string,
    Element as E,
    )

class Plugin(PluginInterface):
    """
    For more information see https://slash.readthedocs.org/en/master/builtin_plugins.html#xunit
    """

    _start_time = datetime.datetime.now()

    def get_name(self):
        return "xunit"

    def get_default_config(self):
        return {
            "filename": "testsuite.xml" // Cmdline(arg="--xunit-filename") // Doc('Name of XML xUnit file to create'),
        }

    def session_start(self):
        self._start_time = datetime.datetime.now()

    def _add_error(self, parent, errortype, error):
        exc_type, exc_value, _ = exc_info = sys.exc_info()
        self._add_element(parent, errortype,
                          {'type': exc_type.__name__ if exc_type else errortype, 'message': error.message},
                          text=get_traceback_string(exc_info) if exc_value is not None else None,
        )

    def _add_element(self, parent, tag, attrib, text=None):
        element = E(tag, attrib)
        if text is not None:
            element.text = text
        parent.append(element)

    def _get_test_case_element(self, test):
        return E('testcase', dict(name=str(test), classname="{}.{}".format(test.__class__.__module__, test.__class__.__name__), time="0"))

    def session_end(self):

        if config.root.parallel.worker_id is not None:
            return

        e = E('testsuite', {
            "name": "slash-suite",
            "hostname": socket.getfqdn(),
            "timestamp": self._start_time.isoformat().rsplit(".", 1)[0],
            "time": "0",
            "tests": str(context.session.results.get_num_results()),
            "errors": str(context.session.results.get_num_errors()),
            "failures": str(context.session.results.get_num_failures()),
            "skipped": str(context.session.results.get_num_skipped()),
        })
        self._add_errors(e, context.session.results.global_result)
        for result in context.session.results.iter_test_results():
            test = E("testcase", {
                "name": result.test_metadata.address,
                "classname": result.test_metadata.class_name or '',
                "time": "0"
            })
            self._add_errors(test, result)

            for skip in result.get_skips():
                self._add_element(test, 'skipped', {'type': skip or ''})

            for detail_name, detail_value in result.details.all().items():
                self._add_element(test, 'detail', {'name': detail_name, 'value': detail_value})

            e.append(test)




        with open(slash_config.root.plugin_config.xunit.filename, "wb") as outfile:
            outfile.write(xml_to_string(e))

    def _add_errors(self, parent, result):
        for error in itertools.chain(result.get_errors(), result.get_failures()):
            self._add_error(parent, 'failure' if error.is_failure() else 'error', error)
